"""Camera related objects and classes."""

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

from pytvpaint import george, utils
from pytvpaint.utils import (
    Refreshable,
    Removable,
    refreshed_property,
    set_as_current,
)

if TYPE_CHECKING:
    from pytvpaint.clip import Clip


class Camera(Refreshable):
    """The Camera class represents the camera in a TVPaint clip.

    There's only one camera in the clip so instantiating multiple objects of this class won't create cameras.
    """

    def __init__(self, clip: Clip, data: george.TVPCamera | None = None) -> None:
        super().__init__()
        self._clip = clip
        self._data: george.TVPCamera = data or george.tv_camera_info_get()
        self._points: list[CameraPoint] = []

    def refresh(self) -> None:
        """Refreshed the camera data."""
        if not self.refresh_on_call and self._data:
            return
        self._data = george.tv_camera_info_get()

    def __repr__(self) -> str:
        """String representation of the camera."""
        return f"Camera()[{self.clip}]"

    @property
    def clip(self) -> Clip:
        """The camera's clip."""
        return self._clip

    def make_current(self) -> None:
        """Makes the parent clip the current one, thereby making sure the correct camera will be set."""
        self._clip.make_current()

    @refreshed_property
    @set_as_current
    def width(self) -> int:
        """The sensor width of the camera."""
        return self._data.width

    @width.setter
    @set_as_current
    def width(self, value: int) -> None:
        george.tv_camera_info_set(width=value, height=self.height)

    @refreshed_property
    @set_as_current
    def height(self) -> int:
        """The sensor height of the camera."""
        return self._data.height

    @height.setter
    @set_as_current
    def height(self, value: int) -> None:
        george.tv_camera_info_set(height=value, width=self.width)

    @refreshed_property
    @set_as_current
    def fps(self) -> float:
        """The framerate of the camera."""
        return self._data.frame_rate

    @fps.setter
    @set_as_current
    def fps(self, value: float) -> None:
        george.tv_camera_info_set(
            self.width,
            self.height,
            self.field_order,
            frame_rate=value,
        )

    @refreshed_property
    @set_as_current
    def pixel_aspect_ratio(self) -> float:
        """The pixel aspect ratio of the camera."""
        return self._data.pixel_aspect_ratio

    @pixel_aspect_ratio.setter
    @set_as_current
    def pixel_aspect_ratio(self, value: float) -> None:
        george.tv_camera_info_set(
            self.width,
            self.height,
            self.field_order,
            self.fps,
            pixel_aspect_ratio=value,
        )

    @refreshed_property
    @set_as_current
    def anti_aliasing(self) -> int:
        """The antialiasing value of the camera."""
        return self._data.anti_aliasing

    @refreshed_property
    @set_as_current
    def field_order(self) -> george.FieldOrder:
        """The field order of the camera."""
        return self._data.field_order

    @field_order.setter
    @set_as_current
    def field_order(self, value: george.FieldOrder) -> None:
        george.tv_camera_info_set(self.width, self.height, field_order=value)

    @set_as_current
    def insert_point(
        self,
        index: int,
        x: int,
        y: int,
        angle: int,
        scale: float,
    ) -> CameraPoint:
        """Insert a new point in the camera path."""
        return CameraPoint.new(self, index, x, y, angle, scale)

    @property
    @set_as_current
    def points(self) -> Iterator[CameraPoint]:
        """Iterator for the `CameraPoint` objects of the camera."""
        points_data = utils.position_generator(
            lambda pos: george.tv_camera_enum_points(pos)
        )
        for index, point_data in enumerate(points_data):
            yield CameraPoint(index, camera=self, data=point_data)

    @set_as_current
    def get_point_data_at(self, position: float) -> george.TVPCameraPoint:
        """Get the points data interpolated at that position (between 0 and 1)."""
        position = max(0.0, min(position, 1.0))
        return george.tv_camera_interpolation(position)

    @set_as_current
    def remove_point(self, index: int) -> None:
        """Remove a point at that index."""
        try:
            point = next(p for i, p in enumerate(self.points) if i == index)
            point.remove()
        except StopIteration:
            pass


class CameraPoint(Removable):
    """A CameraPoint is a point on the camera path.

    You can use them to animate the camera movement.
    """

    def __init__(
        self,
        index: int,
        camera: Camera,
        data: george.TVPCameraPoint | None = None,
    ) -> None:
        super().__init__()
        self._index: int = index
        self._camera: Camera = camera
        self._data = data or george.tv_camera_enum_points(self._index)

    def refresh(self) -> None:
        """Refreshed the camera point data."""
        super().refresh()
        if not self.refresh_on_call and self._data:
            return
        self._data = george.tv_camera_enum_points(self._index)

    def __repr__(self) -> str:
        """String representation of the camera point."""
        return f"CameraPoint({self.camera.clip.name})<index:{self._index}>"

    def __eq__(self, other: object) -> bool:
        """Two camera points are equal if their internal data is the same."""
        if not isinstance(other, CameraPoint):
            return NotImplemented
        self.refresh()
        other.refresh()
        return self._data == other._data

    @property
    def data(self) -> george.TVPCameraPoint:
        """Returns the raw data of the point."""
        return self._data

    @property
    def index(self) -> int:
        """The index of the point in the path."""
        return self._index

    @property
    def camera(self) -> Camera:
        """The camera instance it belongs to."""
        return self._camera

    @refreshed_property
    def x(self) -> float:
        """The x coordinate of the point."""
        return self._data.x

    @x.setter
    def x(self, value: float) -> None:
        current_data = george.tv_camera_enum_points(self.index)
        george.tv_camera_set_point(
            self.index,
            x=value,
            y=current_data.y,
            angle=current_data.angle,
            scale=current_data.scale,
        )

    @refreshed_property
    def y(self) -> float:
        """The y coordinate of the point."""
        return self._data.y

    @y.setter
    def y(self, value: float) -> None:
        current_data = george.tv_camera_enum_points(self.index)
        george.tv_camera_set_point(
            self.index,
            x=current_data.x,
            y=value,
            angle=current_data.angle,
            scale=current_data.scale,
        )

    @refreshed_property
    def angle(self) -> float:
        """The angle of the camera at that point."""
        return self._data.angle

    @angle.setter
    def angle(self, value: float) -> None:
        current_data = george.tv_camera_enum_points(self.index)
        george.tv_camera_set_point(
            self.index,
            x=current_data.x,
            y=current_data.y,
            angle=value,
            scale=current_data.scale,
        )

    @refreshed_property
    def scale(self) -> float:
        """The scale of the camera at that point."""
        return self._data.scale

    @scale.setter
    def scale(self, value: float) -> None:
        current_data = george.tv_camera_enum_points(self.id)
        george.tv_camera_set_point(
            self.id,
            x=current_data.x,
            y=current_data.y,
            angle=current_data.angle,
            scale=value,
        )

    @classmethod
    def new(
        cls,
        camera: Camera,
        index: int,
        x: int,
        y: int,
        angle: int,
        scale: float,
    ) -> CameraPoint:
        """Create a new point and add it to the camera path at that index."""
        george.tv_camera_insert_point(index, x, y, angle, scale)
        return cls(index, camera)

    def remove(self) -> None:
        """Remove the camera point.

        Warning:
            the point instance won't be usable after that call
        """
        george.tv_camera_remove_point(self.index)
        self.mark_removed()



"""Layer related classes."""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import uuid4

from fileseq.filesequence import FileSequence
from fileseq.frameset import FrameSet

from pytvpaint import george, log, utils
from pytvpaint.george.exceptions import GeorgeError
from pytvpaint.utils import (
    Refreshable,
    Removable,
    refreshed_property,
    set_as_current,
)

if TYPE_CHECKING:
    from pytvpaint.clip import Clip
    from pytvpaint.project import Project
    from pytvpaint.scene import Scene


@dataclass
class LayerInstance:
    """A layer instance is a frame where there is a drawing. It only has a start frame.

    Note:
        `LayerInstance` is special because we can't track their position, meaning that if the user move an instance the Python object values won't match.
    """

    layer: Layer
    start: int

    def __post_init__(self) -> None:
        """Checks if the instance exists after init.

        Raises:
            ValueError: if no layer instance found at provided start frame
        """
        try:
            project_start_frame = self.layer.project.start_frame
            george.tv_instance_get_name(self.layer.id, self.start - project_start_frame)
        except GeorgeError:
            raise ValueError(f"There's no instance at frame {self.start}")

    def __eq__(self, other: object) -> bool:
        """Checks if two instances are the same, meaning their start frames are equal."""
        if not isinstance(other, LayerInstance):
            return NotImplemented
        return self.start == other.start

    @property
    def name(self) -> str:
        """Get or set the instance name."""
        self.layer.make_current()
        real_frame = self.start - self.layer.project.start_frame
        return george.tv_instance_get_name(self.layer.id, real_frame)

    @name.setter
    def name(self, value: str) -> None:
        self.layer.make_current()
        real_frame = self.start - self.layer.project.start_frame
        george.tv_instance_set_name(self.layer.id, real_frame, value)

    @property
    def length(self) -> int:
        """Get or set the instance's number of frames or length.

        Raises:
            ValueError: If the length provided is inferior to 1
        """
        self.layer.make_current()
        next_instance = self.next
        end = (next_instance.start - 1) if next_instance else self.layer.end
        return (end - self.start) + 1

    @length.setter
    def length(self, value: int) -> None:
        if value < 1:
            raise ValueError("Instance Length must be at least equal to 1")
        self.layer.make_current()
        real_start = self.start - self.layer.project.start_frame
        george.tv_exposure_set(real_start, value)

    @property
    def end(self) -> int:
        """Get or set the instance's end frame.

        Raises:
            ValueError: If the end frame provided is inferior to the instance's start frame
        """
        return self.start + (self.length - 1)

    @end.setter
    def end(self, value: int) -> None:
        if value < self.start:
            raise ValueError(
                f"End must be equal to or superior to instance start ({self.start})"
            )

        new_length = (value - self.start) + 1
        self.length = new_length

    def split(self, at_frame: int) -> LayerInstance:
        """Split the instance into two instances at the given frame.

        Args:
            at_frame: the frame where the split will occur

        Raises:
            ValueError: If `at_frame` is superior to the instance's end frame

        Returns:
            LayerInstance: the new layer instance
        """
        if at_frame > self.end:
            raise ValueError(
                f"`at_frame` must be in range of the instance's start-end ({self.start}-{self.end})"
            )

        self.layer.make_current()
        real_frame = at_frame - self.layer.project.start_frame
        george.tv_exposure_break(real_frame)

        return LayerInstance(self.layer, at_frame)

    def duplicate(
        self, direction: george.InsertDirection = george.InsertDirection.AFTER
    ) -> None:
        """Duplicate the instance and insert it in the given direction."""
        self.layer.make_current()

        # tvp won't insert images if the insert frame is the same as the instance start, let's move it
        move_frame = self.layer.clip.current_frame
        if move_frame == self.start and self.layer.start != self.start:
            move_frame = self.layer.start
        else:
            move_frame = self.layer.end + 1

        with utils.restore_current_frame(self.layer.clip, move_frame):
            self.copy()
            at_frame = (
                self.end if direction == george.InsertDirection.AFTER else self.start
            )
            self.paste(at_frame=at_frame)

    def cut(self) -> None:
        """Cut all the frames/images/exposures of the instance and store them in the image buffer."""
        self.layer.make_current()
        self.select()
        self.layer.cut_selection()

    def copy(self) -> None:
        """Copy all the frames/images/exposures of the instance and store them in the image buffer."""
        self.layer.make_current()
        self.select()
        self.layer.copy_selection()

    def paste(self, at_frame: int | None) -> None:
        """Paste all the frames/images/exposures stored in the image buffer to the current instance at the given frame.

        Args:
            at_frame: the frame where the stored frames will be pasted. Default is the current frame
        """
        at_frame = at_frame if at_frame is not None else self.layer.clip.current_frame

        self.layer.make_current()
        with utils.restore_current_frame(self.layer.clip, at_frame):
            self.layer.paste_selection()

    def select(self) -> None:
        """Select all frames in this instance."""
        self.layer.select_frames(self.start, (self.length - 1))

    @property
    def next(self) -> LayerInstance | None:
        """Returns the next instance.

        Returns:
            the next instance or None if at the end of the layer
        """
        self.layer.make_current()
        with utils.restore_current_frame(self.layer.clip, self.start):
            next_frame = george.tv_exposure_next()

        next_frame += self.layer.project.start_frame
        if next_frame > self.layer.end:
            return None

        next_instance = LayerInstance(self.layer, next_frame)

        if next_instance == self:
            return None
        return next_instance

    @property
    def previous(self) -> LayerInstance | None:
        """Get the previous instance.

        Returns:
            the previous instance, None if there isn't
        """
        self.layer.make_current()
        with utils.restore_current_frame(self.layer.clip, self.start):
            prev_frame = george.tv_exposure_prev()

        prev_frame += self.layer.project.start_frame
        prev_frame = max(self.layer.start, prev_frame)

        prev_instance = LayerInstance(self.layer, prev_frame)

        if prev_instance == self:
            return None
        return prev_instance


class LayerColor(Refreshable):
    """The color of a layer identified by an index. Layer colors are specific to a clip."""

    def __init__(
        self,
        color_index: int,
        clip: Clip | None = None,
    ) -> None:
        """Construct a LayerColor from an index and a clip (if None it gets the current clip)."""
        from pytvpaint.clip import Clip

        super().__init__()
        self._index = color_index
        self._clip = clip or Clip.current_clip()
        self._data = george.tv_layer_color_get_color(self.clip.id, self._index)

    def refresh(self) -> None:
        """Refreshes the layer color data."""
        if not self.refresh_on_call and self._data:
            return
        self._data = george.tv_layer_color_get_color(self._clip.id, self._index)

    def __repr__(self) -> str:
        """Returns the string representation of LayerColor."""
        return f"LayerColor({self.name})<index:{self.index}>"

    def __eq__(self, other: object) -> bool:
        """Two LayerColor objects are equal if their index are the same."""
        if not isinstance(other, LayerColor):
            return NotImplemented
        return self.index == other.index

    @property
    def index(self) -> int:
        """The layer color index."""
        return self._index

    @property
    def clip(self) -> Clip:
        """The layer color clip."""
        return self._clip

    @refreshed_property
    def name(self) -> str:
        """The name of the color."""
        return self._data.name

    @name.setter
    def name(self, value: str) -> None:
        """Set the name of the color."""
        clip_layer_color_names = (color.name for color in self.clip.layer_colors)
        value = utils.get_unique_name(clip_layer_color_names, value)
        george.tv_layer_color_set_color(self.clip.id, self.index, self.color, value)

    @refreshed_property
    def color(self) -> george.RGBColor:
        """Get the color value."""
        return george.RGBColor(
            r=self._data.color_r,
            g=self._data.color_g,
            b=self._data.color_b,
        )

    @color.setter
    def color(self, value: george.RGBColor) -> None:
        """Set the color value."""
        george.tv_layer_color_set_color(self.clip.id, self.index, value)

    @property
    def is_visible(self) -> bool:
        """Get the visibility of the color index."""
        self.clip.make_current()
        return george.tv_layer_color_visible(self.index)

    def lock_layers(self, lock: bool) -> None:
        """Lock or unlock all layers with this color."""
        self.clip.make_current()
        if lock:
            george.tv_layer_color_lock(self.index)
        else:
            george.tv_layer_color_unlock(self.index)

    def show_layers(
        self,
        show: bool,
        mode: george.LayerColorDisplayOpt = george.LayerColorDisplayOpt.DISPLAY,
    ) -> None:
        """Show or hide layers with this color.

        Args:
            show: whether to show the layers using this color or not
            mode: the display mode. Defaults to george.LayerColorDisplayOpt.DISPLAY.
        """
        self.clip.make_current()
        if show:
            george.tv_layer_color_show(mode, self.index)
        else:
            george.tv_layer_color_hide(mode, self.index)

    def select_layers(self, select: bool) -> None:
        """Select or unselect layers with this color."""
        self.clip.make_current()
        if select:
            george.tv_layer_color_select(self.index)
        else:
            george.tv_layer_color_unselect(self.index)


class Layer(Removable):
    """A Layer is inside a clip and contains drawings."""

    def __init__(self, layer_id: int, clip: Clip | None = None) -> None:
        from pytvpaint.clip import Clip

        super().__init__()
        self._id = layer_id
        self._clip = clip or Clip.current_clip()
        self._data = george.tv_layer_info(self.id)

    def refresh(self) -> None:
        """Refreshes the layer data."""
        super().refresh()
        if not self.refresh_on_call and self._data:
            return
        try:
            self._data = george.tv_layer_info(self._id)
        except GeorgeError:
            self.mark_removed()
            self.refresh()

    def __repr__(self) -> str:
        """The string representation of the layer."""
        return f"Layer({self.name})<id:{self.id}>"

    def __eq__(self, other: object) -> bool:
        """Two layers are equal if their id is the same."""
        if not isinstance(other, Layer):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> int:
        """The layers unique identifier.

        Warning:
            layer ids are not persistent across project load/close
        """
        return self._id

    @property
    def project(self) -> Project:
        """The project containing this layer."""
        return self._clip.project

    @property
    def scene(self) -> Scene:
        """The scene containing this layer."""
        return self._clip.scene

    @property
    def clip(self) -> Clip:
        """The clip containing this layer."""
        return self._clip

    @property
    def position(self) -> int:
        """The position in the layer stack.

        Note:
            layer positions start at 0
        """
        return george.tv_layer_get_pos(self.id)

    @position.setter
    def position(self, value: int) -> None:
        """Moves the layer to the provided position.

        Note:
            This function fixes the issues with positions not been set correctly by TVPaint when value is superior to 0
        """
        if self.position == value:
            return

        value = max(0, value)
        # TVPaint will always set the position at (value - 1) if value is superior to 0, so we need to add +1 in
        #  that case to set the position correctly, I don't know why it works this way, but it honestly makes no sense
        if value != 0:
            value += 1

        self.make_current()
        george.tv_layer_move(value)

    @refreshed_property
    def name(self) -> str:
        """The layer name."""
        return self._data.name

    @name.setter
    @set_as_current
    def name(self, value: str) -> None:
        """Set the layer name.

        Note:
            it uses `get_unique_name` to find a unique layer name across all the layers in the clip
        """
        if value == self.name:
            return
        value = utils.get_unique_name(self.clip.layer_names, value)
        george.tv_layer_rename(self.id, value)

    @refreshed_property
    def layer_type(self) -> george.LayerType:
        """The layer type."""
        return self._data.type

    @property
    def opacity(self) -> int:
        """Get the layer opacity value.

        Note:
            In George, this is called density, we renamed it to `opacity` as it seems more appropriate
        """
        return george.tv_layer_density_get()

    @opacity.setter
    @set_as_current
    def opacity(self, value: int) -> None:
        """Set the layers opacity value (between 0 and 100)."""
        value = max(0, min(value, 100))
        george.tv_layer_density_set(value)

    @refreshed_property
    def start(self) -> int:
        """The layer start frame according to the project's start frame."""
        return self._data.first_frame + self.project.start_frame

    @refreshed_property
    def end(self) -> int:
        """The layer end frame according to the project's start frame."""
        return self._data.last_frame + self.project.start_frame

    @property
    def color(self) -> LayerColor:
        """Get the layer color."""
        color_index = george.tv_layer_color_get(self.id)
        return next(c for i, c in enumerate(self.clip.layer_colors) if i == color_index)

    @color.setter
    def color(self, color: LayerColor) -> None:
        """Set the layer color."""
        george.tv_layer_color_set(self.id, color.index)

    @property
    def is_current(self) -> bool:
        """Returns True if the layer is the current one in the clip."""
        return self.id == self.current_layer_id()

    def make_current(self) -> None:
        """Make the layer current, it also makes the clip current."""
        if self.is_current:
            return
        if self.clip:
            self.clip.make_current()
        george.tv_layer_set(self.id)

    @property
    def is_selected(self) -> bool:
        """Returns True if the layer is selected."""
        return george.tv_layer_selection_get(self.id)

    @is_selected.setter
    def is_selected(self, value: bool) -> None:
        """Select or deselect the layer."""
        george.tv_layer_selection_set(self.id, new_state=value)

    @property
    def is_visible(self) -> bool:
        """Returns True if the layer is visible."""
        return george.tv_layer_display_get(self.id)

    @is_visible.setter
    def is_visible(self, value: bool) -> None:
        """Set the visibility state of the layer."""
        george.tv_layer_display_set(self.id, new_state=value)

    @property
    def is_locked(self) -> bool:
        """Returns True if the layer is locked."""
        return george.tv_layer_lock_get(self.id)

    @is_locked.setter
    def is_locked(self, value: bool) -> None:
        """Lock or unlock the layer."""
        george.tv_layer_lock_set(self.id, new_state=value)

    @property
    def is_collapsed(self) -> bool:
        """Returns True if the layer is collapsed."""
        return george.tv_layer_collapse_get(self.id)

    @is_collapsed.setter
    def is_collapsed(self, value: bool) -> None:
        """Collapse or uncollapse the layer."""
        george.tv_layer_collapse_set(self.id, new_state=value)

    @property
    def blending_mode(self) -> george.BlendingMode:
        """Get the layer blending mode value."""
        return george.tv_layer_blending_mode_get(self.id)

    @blending_mode.setter
    def blending_mode(self, value: george.BlendingMode) -> None:
        """Set the layer blending mode value."""
        george.tv_layer_blending_mode_set(self.id, mode=value)

    @property
    def stencil(self) -> george.StencilMode:
        """Get the layer stencil mode value."""
        return george.tv_layer_stencil_get(self.id)

    @stencil.setter
    def stencil(self, mode: george.StencilMode) -> None:
        """Set the layer stencil mode value."""
        george.tv_layer_stencil_set(self.id, mode=mode)

    @property
    def thumbnails_visible(self) -> bool:
        """Returns True if thumbnails are shown on that layer."""
        return george.tv_layer_show_thumbnails_get(self.id)

    @thumbnails_visible.setter
    def thumbnails_visible(self, value: bool) -> None:
        """Show or hide the thumbnails for that layer."""
        george.tv_layer_show_thumbnails_set(self.id, value)

    @property
    def auto_break_instance(self) -> bool:
        """Get the auto break instance value."""
        return george.tv_layer_auto_break_instance_get(self.id)

    @auto_break_instance.setter
    def auto_break_instance(self, value: bool) -> None:
        """Set the auto break instance value.

        Raises:
            ValueError: the layer is not an animation layer
        """
        if value and not self.is_anim_layer:
            msg = "Can't set auto break instance because it's not an animation layer"
            raise ValueError(msg)
        george.tv_layer_auto_break_instance_set(self.id, value)

    @property
    def auto_create_instance(self) -> bool:
        """Get the auto create instance value."""
        return george.tv_layer_auto_create_instance_get(self.id)

    @auto_create_instance.setter
    def auto_create_instance(self, value: bool) -> None:
        """Set the auto create instance value."""
        george.tv_layer_auto_create_instance_set(self.id, value)

    @property
    def pre_behavior(self) -> george.LayerBehavior:
        """Get the pre-behavior value."""
        return george.tv_layer_pre_behavior_get(self.id)

    @pre_behavior.setter
    def pre_behavior(self, value: george.LayerBehavior) -> None:
        """Set the pre-behavior value."""
        george.tv_layer_pre_behavior_set(self.id, value)

    @property
    def post_behavior(self) -> george.LayerBehavior:
        """Get the post-behavior value."""
        return george.tv_layer_post_behavior_get(self.id)

    @post_behavior.setter
    def post_behavior(self, value: george.LayerBehavior) -> None:
        """Set the post-behavior value."""
        george.tv_layer_post_behavior_set(self.id, value)

    @property
    def is_position_locked(self) -> bool:
        """Returns True if the layer position is locked."""
        return george.tv_layer_lock_position_get(self.id)

    @is_position_locked.setter
    def is_position_locked(self, value: bool) -> None:
        """Lock or unlock the layer position."""
        george.tv_layer_lock_position_set(self.id, value)

    @property
    def preserve_transparency(self) -> george.LayerTransparency:
        """Get the preserve transparency value."""
        return george.tv_preserve_get()

    @preserve_transparency.setter
    @set_as_current
    def preserve_transparency(self, value: george.LayerTransparency) -> None:
        """Set the preserve transparency value."""
        george.tv_preserve_set(value)

    @set_as_current
    def convert_to_anim_layer(self) -> None:
        """Converts the layer to an animation layer."""
        george.tv_layer_anim(self.id)

    @property
    def is_anim_layer(self) -> bool:
        """Returns True if the layer is an animation layer."""
        return self.layer_type == george.LayerType.SEQUENCE

    def load_dependencies(self) -> None:
        """Load all dependencies of the layer in memory."""
        george.tv_layer_load_dependencies(self.id)

    @staticmethod
    def current_layer_id() -> int:
        """Returns the current layer id."""
        return george.tv_layer_current_id()

    @classmethod
    def current_layer(cls) -> Layer:
        """Returns the current layer object."""
        from pytvpaint.clip import Clip

        return cls(layer_id=cls.current_layer_id(), clip=Clip.current_clip())

    @set_as_current
    def shift(self, new_start: int) -> None:
        """Move the layer to a new frame."""
        george.tv_layer_shift(self.id, new_start - self.project.start_frame)

    @set_as_current
    def merge(
        self,
        layer: Layer,
        blending_mode: george.BlendingMode,
        stamp: bool = False,
        erase: bool = False,
        keep_color_grp: bool = True,
        keep_img_mark: bool = True,
        keep_instance_name: bool = True,
    ) -> None:
        """Merge this layer with the given one.

        Args:
            layer: the layer to merge with
            blending_mode: the blending mode to use
            stamp: Use stamp mode
            erase: Remove the source layer
            keep_color_grp: Keep the color group
            keep_img_mark: Keep the image mark
            keep_instance_name: Keep the instance name
        """
        george.tv_layer_merge(
            layer.id,
            blending_mode,
            stamp,
            erase,
            keep_color_grp,
            keep_img_mark,
            keep_instance_name,
        )

    @staticmethod
    def merge_all(
        keep_color_grp: bool = True,
        keep_img_mark: bool = True,
        keep_instance_name: bool = True,
    ) -> None:
        """Merge all the layers in the stack.

        Args:
            keep_color_grp: Keep the color group
            keep_img_mark: Keep the image mark
            keep_instance_name: Keep the instance name
        """
        george.tv_layer_merge_all(keep_color_grp, keep_img_mark, keep_instance_name)

    @staticmethod
    @george.undoable
    def new(
        name: str,
        clip: Clip | None = None,
        color: LayerColor | None = None,
    ) -> Layer:
        """Create a new layer.

        Args:
            name: the name of the new layer
            clip: the parent clip
            color: the layer color

        Returns:
            Layer: the new layer

        Note:
            The layer name is checked against all other layers to have a unique name using `get_unique_name`.
            This can take a while if you have a lot of layers.
        """
        from pytvpaint.clip import Clip

        clip = clip or Clip.current_clip()
        clip.make_current()

        name = utils.get_unique_name(clip.layer_names, name)
        layer_id = george.tv_layer_create(name)

        layer = Layer(layer_id=layer_id, clip=clip)

        if color:
            layer.color = color

        return layer

    @classmethod
    @george.undoable
    def new_anim_layer(
        cls,
        name: str,
        clip: Clip | None = None,
        color: LayerColor | None = None,
    ) -> Layer:
        """Create a new animation layer.

        Args:
            name: the name of the new layer
            clip: the parent clip
            color: the layer color

        Returns:
            Layer: the new animation layer

        Note:
            It activates the thumbnail visibility
        """
        layer = cls.new(name, clip, color)
        layer.convert_to_anim_layer()
        layer.thumbnails_visible = True
        return layer

    @classmethod
    @george.undoable
    def new_background_layer(
        cls,
        name: str,
        clip: Clip | None = None,
        color: LayerColor | None = None,
        image: Path | str | None = None,
        stretch: bool = False,
    ) -> Layer:
        """Create a new background layer with hold as pre- and post-behavior.

        Args:
            name: the name of the new layer
            clip: the parent clip
            color: the layer color
            image: the background image to load
            stretch: whether to stretch the image to fit the view

        Returns:
            Layer: the new animation layer
        """
        from pytvpaint.clip import Clip

        clip = clip or Clip.current_clip()
        layer = cls.new(name, clip, color)
        layer.pre_behavior = george.LayerBehavior.HOLD
        layer.post_behavior = george.LayerBehavior.HOLD
        layer.thumbnails_visible = True

        image = Path(image or "")
        if image.is_file():
            layer.convert_to_anim_layer()
            layer.load_image(image, frame=clip.start, stretch=stretch)

        return layer

    @set_as_current
    @george.undoable
    def duplicate(self, name: str) -> Layer:
        """Duplicate this layer.

        Args:
            name: the duplicated layer name

        Returns:
            Layer: the duplicated layer
        """
        name = utils.get_unique_name(self.clip.layer_names, name)
        layer_id = george.tv_layer_duplicate(name)

        return Layer(layer_id=layer_id, clip=self.clip)

    def remove(self) -> None:
        """Remove the layer from the clip.

        Warning:
            The current instance won't be usable after this call since it will be mark removed.
        """
        self.clip.make_current()
        self.is_locked = False
        george.tv_layer_kill(self.id)
        self.mark_removed()

    @set_as_current
    def render(
        self,
        output_path: Path | str | FileSequence,
        start: int | None = None,
        end: int | None = None,
        use_camera: bool = False,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Render the layer to a single frame or frame sequence or movie.

        Args:
            output_path: a single file or file sequence pattern
            start: the start frame to render the layer's start if None. Defaults to None.
            end: the end frame to render or the layer's end if None. Defaults to None.
            use_camera: use the camera for rendering, otherwise render the whole canvas. Defaults to False.
            alpha_mode: the alpha mode for rendering. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the background mode for rendering. Defaults to None.
            format_opts: custom format options. Defaults to None.

        Raises:
            ValueError: if requested range (start-end) not in clip range/bounds
            ValueError: if output is a movie
            FileNotFoundError: if the render failed and no files were found on disk or missing frames

        Note:
            This functions uses the layer's range as a basis (start-end). This  is different from a project range, which
            uses the project timeline. For more details on the differences in frame ranges and the timeline in TVPaint,
            please check the `Usage/Rendering` section of the documentation.

        Warning:
            Even tough pytvpaint does a pretty good job of correcting the frame ranges for rendering, we're still
            encountering some weird edge cases where TVPaint will consider the range invalid for seemingly no reason.
        """
        start = self.start if start is None else start
        end = self.end if end is None else end
        self.clip.render(
            output_path=output_path,
            start=start,
            end=end,
            use_camera=use_camera,
            layer_selection=[self],
            alpha_mode=alpha_mode,
            background_mode=background_mode,
            format_opts=format_opts,
        )

    @set_as_current
    def render_frame(
        self,
        export_path: Path | str,
        frame: int | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = george.BackgroundMode.NONE,
        format_opts: list[str] | None = None,
    ) -> Path:
        """Render a frame from the layer.

        Args:
            export_path: the frame export path (the extension determines the output format)
            frame: the frame to render or the current frame if None. Defaults to None.
            alpha_mode: the render alpha mode
            background_mode: the render background mode
            format_opts: custom output format options to pass when rendering

        Raises:
            FileNotFoundError: if the render failed or output not found on disk

        Returns:
            Path: render output path
        """
        export_path = Path(export_path)
        save_format = george.SaveFormat.from_extension(export_path.suffix)
        export_path.parent.mkdir(parents=True, exist_ok=True)

        frame = frame or self.clip.current_frame
        self.clip.current_frame = frame

        with utils.render_context(
            alpha_mode,
            background_mode,
            save_format,
            format_opts,
            layer_selection=[self],
        ):
            george.tv_save_image(export_path)

        if not export_path.exists():
            raise FileNotFoundError(
                f"Could not find rendered image ({frame}) at : {export_path.as_posix()}"
            )

        return export_path

    @set_as_current
    def render_instances(
        self,
        export_path: Path | str | FileSequence,
        start: int | None = None,
        end: int | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> FileSequence:
        """Render all layer instances in the provided range for the current layer.

        Args:
            export_path: the export path (the extension determines the output format)
            start: the start frame to render the layer's start if None. Defaults to None.
            end: the end frame to render or the layer's end if None. Defaults to None.
            alpha_mode: the render alpha mode
            background_mode: the render background mode
            format_opts: custom output format options to pass when rendering

        Raises:
            ValueError: if requested range (start-end) not in layer range/bounds
            ValueError: if output is a movie
            FileNotFoundError: if the render failed or output not found on disk

        Returns:
            FileSequence: instances output sequence
        """
        file_sequence, start, end, is_sequence, is_image = utils.handle_output_range(
            export_path, self.start, self.end, start, end
        )

        if start < self.start or end > self.end:
            raise ValueError(
                f"Render ({start}-{end}) not in clip range ({(self.start, self.end)})"
            )
        if not is_image:
            raise ValueError(
                f"Video formats ({file_sequence.extension()}) are not supported for instance rendering !"
            )

        # render to output
        frames = []
        for layer_instance in self.instances:
            cur_frame = layer_instance.start
            instance_output = Path(file_sequence.frame(cur_frame))
            self.render_frame(
                instance_output, cur_frame, alpha_mode, background_mode, format_opts
            )
            frames.append(str(cur_frame))

        file_sequence.setFrameSet(FrameSet(",".join(frames)))
        return file_sequence

    @set_as_current
    def load_image(
        self, image_path: str | Path, frame: int | None = None, stretch: bool = False
    ) -> None:
        """Load an image in the current layer at a given frame.

        Args:
            image_path: path to the image to load
            frame: the frame where the image will be loaded, if none provided, image will be loaded at current frame
            stretch: whether to stretch the image to fit the view

        Raises:
            FileNotFoundError: if the file doesn't exist at provided path
        """
        image_path = Path(image_path)
        if not image_path.exists():
            raise FileNotFoundError(f"Image not found at : {image_path}")

        frame = frame or self.clip.current_frame
        with utils.restore_current_frame(self.clip, frame):
            # if no instance at the specified frame, then create a new one
            if not self.get_instance(frame):
                self.add_instance(frame)

            george.tv_load_image(image_path.as_posix(), stretch)

    def get_mark_color(self, frame: int) -> LayerColor | None:
        """Get the mark color at a specific frame.

        Args:
            frame: frame with a mark

        Returns:
            LayerColor | None: the layer color if there was a mark
        """
        frame = frame - self.project.start_frame
        color_index = george.tv_layer_mark_get(self.id, frame)
        if not color_index:
            return None

        return self.clip.get_layer_color(by_index=color_index)

    def add_mark(self, frame: int, color: LayerColor) -> None:
        """Add a mark to a frame.

        Args:
            frame: frame to put a mark on
            color: the color index

        Raises:
            TypeError: if the layer is not an animation layer
        """
        if not self.is_anim_layer:
            raise TypeError(
                f"Can't add a mark because this is not an animation layer ({self})"
            )
        frame = frame - self.project.start_frame
        george.tv_layer_mark_set(self.id, frame, color.index)

    def remove_mark(self, frame: int) -> None:
        """Remove a mark.

        Args:
            frame: a frame number with a mark
        """
        # Setting it at 0 clears the mark
        self.add_mark(frame, LayerColor(0, self.clip))

    @property
    def marks(self) -> Iterator[tuple[int, LayerColor]]:
        """Iterator over the layer marks including the frame and the color.

        Yields:
            frame (int): the mark frame
            color (LayerColor): the mark color
        """
        for frame in range(self.start, self.end + 1):
            layer_color = self.get_mark_color(frame)
            if not layer_color:
                continue
            yield (frame, layer_color)

    def clear_marks(self) -> None:
        """Clear all the marks in the layer."""
        for frame, _ in self.marks:
            self.remove_mark(frame)

    @set_as_current
    def select_frames(self, start: int, end: int) -> None:
        """Select the frames from a start and count.

        Args:
            start: the selection start frame
            end: the selected end frame
        """
        if not self.is_anim_layer:
            log.warning(
                "Selection may display weird behaviour when applied to a non animation layer"
            )
        frame_count = (end - start) + 1
        george.tv_layer_select(start - self.clip.start, frame_count)

    @set_as_current
    def select_all_frames(self) -> None:
        """Select all frames in the layer."""
        frame, frame_count = george.tv_layer_select_info(full=True)
        george.tv_layer_select(frame, frame_count)

    @set_as_current
    def clear_selection(self) -> None:
        """Clear frame selection in the layer."""
        # selecting frames after the layer's end frame will result in a empty selection, thereby clearing the selection
        george.tv_layer_select(self.end + 1, 0)

    @property
    def selected_frames(self) -> list[int]:
        """Get the list of selected frames.

        Returns:
            Array of selected frame numbers
        """
        start, count = george.tv_layer_select_info(full=False)
        return [(start + self.clip.start + offset) for offset in range(count)]

    @set_as_current
    def cut_selection(self) -> None:
        """Cut the selected instances."""
        george.tv_layer_cut()

    @set_as_current
    def copy_selection(self) -> None:
        """Copy the selected instances."""
        george.tv_layer_copy()

    @set_as_current
    def paste_selection(self) -> None:
        """Paste the previously copied instances."""
        george.tv_layer_paste()

    @refreshed_property
    @set_as_current
    def instances(self) -> Iterator[LayerInstance]:
        """Iterates over the layer instances.

        Yields:
            each LayerInstance present in the layer
        """
        # instances start at layer start
        current_instance = LayerInstance(self, self.start)

        while True:
            yield current_instance

            nex_instance = current_instance.next
            if nex_instance is None:
                break

            current_instance = nex_instance

    def get_instance(self, frame: int, strict: bool = False) -> LayerInstance | None:
        """Get the instance at that frame.

        Args:
            frame: the instance frame
            strict: True will only return Instance if the given frame is the start of the instance. Default is False

        Returns:
            the instance if found else None
        """
        for layer_instance in self.instances:
            if strict:
                if layer_instance.start != frame:
                    continue
                return layer_instance

            if not (layer_instance.start <= frame <= layer_instance.end):
                continue
            return layer_instance

        return None

    def get_instances(self, from_frame: int, to_frame: int) -> Iterator[LayerInstance]:
        """Iterates over the layer instances and returns the one in the range (from_frame-to_frame).

        Yields:
            each LayerInstance in the range (from_frame-to_frame)
        """
        for layer_instance in self.instances:
            if layer_instance.end < from_frame:
                continue
            if from_frame <= layer_instance.start <= to_frame:
                yield layer_instance
            if layer_instance.start > to_frame:
                break

    def add_instance(
        self,
        start: int | None = None,
        nb_frames: int = 1,
        direction: george.InsertDirection | None = None,
        split: bool = False,
    ) -> LayerInstance:
        """Crates a new instance.

        Args:
            start: start frame. Defaults to clip current frame if none provided
            nb_frames: number of frames in the new instance. Default is 1, this is the total number of frames created.
            direction: direction where new frames will be added/inserted
            split: True to make each added frame a new image

        Raises:
            TypeError: if the layer is not an animation layer
            ValueError: if the number of frames `nb_frames` is inferior or equal to 0
            ValueError: if an instance already exists at the given range (start + nb_frames)

        Returns:
            LayerInstance: new layer instance
        """
        if not self.is_anim_layer:
            raise TypeError("The layer needs to be an animation layer")

        if nb_frames <= 0:
            raise ValueError("Instance number of frames must be at least 1")

        if start and self.get_instance(start):
            raise ValueError(
                "An instance already exists at the designated frame range. "
                "Edit or delete it before adding a new one."
            )

        start = start if start is not None else self.clip.current_frame
        self.clip.make_current()

        temp_layer = Layer.new_anim_layer(str(uuid4()))
        temp_layer.make_current()

        with utils.restore_current_frame(self.clip, 1):
            if nb_frames > 1:
                if split:
                    george.tv_layer_insert_image(count=nb_frames, direction=direction)
                else:
                    layer_instance = next(temp_layer.instances)
                    layer_instance.length = nb_frames

            temp_layer.select_all_frames()
            temp_layer.copy_selection()
            self.clip.current_frame = start
            self.make_current()
            self.paste_selection()
            temp_layer.remove()

        return LayerInstance(self, start)

    def rename_instances(
        self,
        mode: george.InstanceNamingMode,
        prefix: str | None = None,
        suffix: str | None = None,
        process: george.InstanceNamingProcess | None = None,
    ) -> None:
        """Rename all the instances.

        Args:
            mode: the instance renaming mode
            prefix: the prefix to add to each name
            suffix: the suffix to add to each name
            process: the instance naming process
        """
        george.tv_instance_name(self.id, mode, prefix, suffix, process)



"""Utility functions and classes which are not specific to anything else in the codebase."""

from __future__ import annotations

import contextlib
import re
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, Iterator
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    TypeVar,
    cast,
)

from fileseq.filesequence import FileSequence
from fileseq.frameset import FrameSet
from typing_extensions import ParamSpec, Protocol

from pytvpaint import george
from pytvpaint.george.exceptions import GeorgeError

if TYPE_CHECKING:
    from pytvpaint.layer import Layer


class _CanRefresh(Protocol):
    def refresh(self) -> None: ...


if TYPE_CHECKING:
    refreshed_property = property
else:

    class RefreshedProperty(property):
        """Custom property that calls .refresh() before getting the actual value."""

        def __get__(self, __obj: _CanRefresh, __type: type | None = None) -> Any:
            """Calls .refresh() on the object before getting the value."""
            __obj.refresh()
            return super().__get__(__obj, __type)

    refreshed_property = RefreshedProperty


class Refreshable(ABC):
    """Abstract class that denotes an object that have data that can be refreshed (a TVPaint project for example)."""

    def __init__(self) -> None:
        self.refresh_on_call = True

    @abstractmethod
    def refresh(self) -> None:
        """Refreshes the object data."""
        raise NotImplementedError("Function refresh() needs to be implemented")


class Removable(Refreshable):
    """Abstract class that denotes an object that can be removed from TVPaint (a Layer for example)."""

    def __init__(self) -> None:
        super().__init__()
        self._is_removed: bool = False

    def __getattribute__(self, name: str) -> Any:
        """For each attribute access, we check if the object was marked removed."""
        if not name.startswith("_") and self._is_removed:
            raise ValueError(f"{self.__class__.__name__} has been removed!")

        return super().__getattribute__(name)

    def refresh(self) -> None:
        """Does a refresh of the object data.

        Raises:
            ValueError: if the object has been mark removed
        """
        if self._is_removed:
            raise ValueError(f"{self.__class__.__name__} has been removed!")

    @property
    def is_removed(self) -> bool:
        """Checks if the object is removed by trying to refresh its data.

        Returns:
            bool: whether if it was removed or not
        """
        self._is_removed = False
        with contextlib.suppress(Exception):
            self.refresh()
            self._is_removed = True
        return self._is_removed

    @abstractmethod
    def remove(self) -> None:
        """Removes the object in TVPaint."""
        raise NotImplementedError("Function refresh() needs to be implemented")

    def mark_removed(self) -> None:
        """Marks the object as removed and is therefor not usable."""
        self._is_removed = True


class Renderable(ABC):
    """Abstract class that denotes an object that can be removed from TVPaint (a Layer for example)."""

    def __init__(self) -> None:
        super().__init__()

    @property
    @abstractmethod
    def current_frame(self) -> int:
        """Gives the current frame."""
        pass

    @current_frame.setter
    @abstractmethod
    def current_frame(self, frame: int) -> None:
        """Set the current frame."""
        pass

    def _get_real_range(self, start: int, end: int) -> tuple[int, int]:
        """Removes the object in TVPaint."""
        raise NotImplementedError("Function refresh() needs to be implemented")

    def _validate_range(self, start: int, end: int) -> None:
        """Raises an exception if given range is invalid."""
        raise NotImplementedError("Function refresh() needs to be implemented")

    def _render(
        self,
        output_path: Path | str | FileSequence,
        default_start: int,
        default_end: int,
        start: int | None = None,
        end: int | None = None,
        use_camera: bool = False,
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        file_sequence, start, end, is_sequence, is_image = handle_output_range(
            output_path, default_start, default_end, start, end
        )
        self._validate_range(start, end)

        origin_start = int(start)
        start, end = self._get_real_range(start, end)
        if not is_image and start == end:
            raise ValueError(
                "TVPaint will not render a movie that contains a single frame"
            )

        # get first frame, tvp doesn't understand vfx padding `#`
        if is_image or file_sequence.padding():
            first_frame = Path(file_sequence.frame(file_sequence.start()))
        else:
            first_frame = Path(str(output_path))
        first_frame.parent.mkdir(exist_ok=True, parents=True)

        save_format = george.SaveFormat.from_extension(
            file_sequence.extension().lower()
        )

        # render to output
        with render_context(
            alpha_mode, background_mode, save_format, format_opts, layer_selection
        ):
            if start == end:
                with restore_current_frame(self, origin_start):
                    george.tv_save_display(first_frame)
            else:
                # not using tv_save_sequence since it doesn't handle camera and would require different range math
                george.tv_project_save_sequence(
                    first_frame,
                    start=start,
                    end=end,
                    use_camera=use_camera,
                )

        # make sure the output exists otherwise raise an error
        if is_sequence:
            # raises error if sequence not found
            found_sequence = FileSequence.findSequenceOnDisk(str(file_sequence))
            frame_set = found_sequence.frameSet()
            file_sequence_frame_set = file_sequence.frameSet()

            if file_sequence_frame_set is None or frame_set is None:
                raise Exception("Should have frame set")

            if not frame_set.issuperset(file_sequence_frame_set):
                # not all frames found
                missing_frames = file_sequence_frame_set.difference(frame_set)
                raise FileNotFoundError(
                    f"Not all frames found, missing frames ({missing_frames}) "
                    f"in sequence : {output_path}"
                )
        else:
            if not first_frame.exists():
                raise FileNotFoundError(
                    f"Could not find output at : {first_frame.as_posix()}"
                )


def get_unique_name(names: Iterable[str], stub: str) -> str:
    """Get a unique name from a list of names and a stub prefix. It does auto increment it.

    Args:
        names (Iterable[str]): existing names
        stub (str): the base name

    Raises:
        ValueError: if the stub is empty

    Returns:
        str: a unique name with the stub prefix
    """
    if not stub:
        raise ValueError("Stub is empty")

    number_re = re.compile(r"(?P<number>\d+)$", re.I)

    stub_without_number = number_re.sub("", stub)
    max_number = 0
    padding_length = 1

    for name in names:
        without_number = number_re.sub("", name)

        if without_number != stub_without_number:
            continue

        res = number_re.search(name)
        number = res.group("number") if res else "1"

        padding_length = max(padding_length, len(number))
        max_number = max(max_number, int(number))

    if max_number == 0:
        return stub

    next_number = max_number + 1
    return f"{stub_without_number}{next_number:0{padding_length}}"


T = TypeVar("T")


def position_generator(
    fn: Callable[[int], T],
    stop_when: type[GeorgeError] = GeorgeError,
) -> Iterator[T]:
    """Utility generator that yields the result of a function according to a position.

    Args:
        fn (Callable[[int], T]): the function to run at each iteration
        stop_when (Type[GeorgeError], optional): exception at which we stop. Defaults to GeorgeError.

    Yields:
        Iterator[T]: an generator of the resulting values
    """
    pos = 0

    while True:
        try:
            yield fn(pos)
        except stop_when:
            break
        pos += 1


class CanMakeCurrent(Protocol):
    """Describes an object that can do `make_current` and has an id."""

    @property
    def id(self) -> str | int:  # noqa: D102
        ...

    def make_current(self) -> None:  # noqa: D102
        ...


# See: https://stackoverflow.com/questions/47060133/python-3-type-hinting-for-decorator
Params = ParamSpec("Params")
ReturnType = TypeVar("ReturnType")


def set_as_current(func: Callable[Params, ReturnType]) -> Callable[Params, ReturnType]:
    """Decorator to apply on object methods.

    Sets the current TVPaint object as 'current'.
    Useful when George functions only apply on the current project, clip, layer or scene.

    Args:
        func (Callable[Params, ReturnType]): the method apply on

    Returns:
        Callable[Params, ReturnType]: the wrapped method
    """

    def wrapper(*args: Params.args, **kwargs: Params.kwargs) -> ReturnType:
        self = cast(CanMakeCurrent, args[0])
        self.make_current()
        return func(*args, **kwargs)

    return wrapper


@contextlib.contextmanager
def render_context(
    alpha_mode: george.AlphaSaveMode | None = None,
    background_mode: george.BackgroundMode | None = None,
    save_format: george.SaveFormat | None = None,
    format_opts: list[str] | None = None,
    layer_selection: list[Layer] | None = None,
) -> Generator[None, None, None]:
    """Context used to do renders in TVPaint.

    It does the following things:

    - Set the alpha mode and save format (with custom options)
    - Hide / Show the given layers (some render functions only render by visibility)
    - Restore the previous values after rendering

    Args:
        alpha_mode: the render alpha save mode
        save_format: the render format to use. Defaults to None.
        background_mode: the render background mode
        format_opts: the custom format options as strings. Defaults to None.
        layer_selection: the layers to render. Defaults to None.
    """
    from pytvpaint.clip import Clip

    # Save the current state
    pre_alpha_save_mode = george.tv_alpha_save_mode_get()
    pre_save_format, pre_save_args = george.tv_save_mode_get()
    pre_background_mode, pre_background_colors = george.tv_background_get()

    # Set the save mode values
    if alpha_mode:
        george.tv_alpha_save_mode_set(alpha_mode)
    if background_mode:
        george.tv_background_set(background_mode)
    if save_format:
        george.tv_save_mode_set(save_format, *(format_opts or []))

    layers_visibility = []
    if layer_selection:
        clip = Clip.current_clip()
        layers_visibility = [(layer, layer.is_visible) for layer in clip.layers]
        # Show and hide the clip layers to render
        for layer, _ in layers_visibility:
            should_be_visible = not layer_selection or layer in layer_selection
            layer.is_visible = should_be_visible

    # Do the render
    yield

    # Restore the previous values
    if alpha_mode:
        george.tv_alpha_save_mode_set(pre_alpha_save_mode)
    if save_format:
        george.tv_save_mode_set(pre_save_format, *pre_save_args)
    if background_mode:
        george.tv_background_set(pre_background_mode, pre_background_colors)

    # Restore the layer visibility
    if layers_visibility:
        for layer, was_visible in layers_visibility:
            layer.is_visible = was_visible


class HasCurrentFrame(Protocol):
    """Class that has a current frame property."""

    @property
    def current_frame(self) -> int:
        """The current frame, clip or project."""
        ...

    @current_frame.setter
    def current_frame(self, value: int) -> None: ...


@contextlib.contextmanager
def restore_current_frame(
    tvp_element: HasCurrentFrame, frame: int
) -> Generator[None, None, None]:
    """Context that temporarily changes the current frame to the one provided and restores it when done.

    Args:
        tvp_element: clip to change
        frame: frame to set. Defaults to None.
    """
    previous_frame = tvp_element.current_frame

    if frame != previous_frame:
        tvp_element.current_frame = frame

    yield

    if tvp_element.current_frame != previous_frame:
        tvp_element.current_frame = previous_frame


class _TVPElement(Protocol):
    @property
    def id(self) -> int | str: ...

    @property
    def name(self) -> str: ...


TVPElementType = TypeVar("TVPElementType", bound=_TVPElement)


def get_tvp_element(
    tvp_elements: Iterator[TVPElementType],
    by_id: int | str | None = None,
    by_name: str | None = None,
    by_path: str | Path | None = None,
) -> TVPElementType | None:
    """Search for a TVPaint element by attributes.

    Args:
        tvp_elements: a collection of TVPaint objects
        by_id: search by id. Defaults to None.
        by_name: search by name, search is case-insensitive. Defaults to None.
        by_path: search by path. Defaults to None.

    Raises:
        ValueError: if bad arguments were given

    Returns:
        TVPElementType | None: the found element
    """
    if by_id is None and by_name is None:
        raise ValueError(
            "At least one of the values (id or name) must be provided, none found !"
        )

    for element in tvp_elements:
        if by_id is not None and element.id != by_id:
            continue
        if by_name is not None and element.name.lower() != by_name.lower():
            continue
        if by_path is not None and getattr(element, "path") != Path(by_path):
            continue
        return element

    return None


def handle_output_range(
    output_path: Path | str | FileSequence,
    default_start: int,
    default_end: int,
    start: int | None = None,
    end: int | None = None,
) -> tuple[FileSequence, int, int, bool, bool]:
    """Handle the different options for output paths and range.

    Whether the user provides a range (start-end) or a filesequence with a range or not, this functions ensures we
    always end up with a valid range to render

    Args:
        output_path: user provided output path
        default_start: the default start to use if none provided or found in the file sequence object
        default_end: the default end to use if none provided or found in the file sequence object
        start: user provided start frame or None
        end: user provided end frame or None

    Returns:
        file_sequence: output path as a FileSequence object
        start: computed start frame
        end: computed end frame
        is_sequence: whether the output is a sequence or not
        is_image: whether the output is an image or not (a movie)
    """
    # we handle all outputs as a FileSequence, makes it a bit easier to handle ranges and padding
    if not isinstance(output_path, FileSequence):
        file_sequence = FileSequence(Path(output_path).as_posix())
    else:
        file_sequence = output_path

    frame_set = file_sequence.frameSet()
    is_image = george.SaveFormat.is_image(file_sequence.extension())

    # if the provided sequence has a range, and we don't, use the sequence range
    if frame_set and len(frame_set) >= 1 and is_image:
        start = start or file_sequence.start()
        end = end or file_sequence.end()

    # check characteristics of file sequence
    fseq_has_range = frame_set and len(frame_set) > 1
    fseq_is_single_image = frame_set and len(frame_set) == 1
    fseq_no_range_padding = not frame_set and file_sequence.padding()
    range_is_seq = start is not None and end is not None and start != end
    range_is_single_image = start is not None and end is not None and start == end

    is_single_image = bool(
        is_image and (fseq_is_single_image or not frame_set) and range_is_single_image
    )
    is_sequence = bool(
        is_image and (fseq_has_range or fseq_no_range_padding or range_is_seq)
    )

    # if no range provided, use clip mark in/out, if none, use clip start/end
    if start is None:
        start = default_start
    if is_single_image and not end:
        end = start
    else:
        if end is None:
            end = default_end

    frame_set = FrameSet(f"{start}-{end}")

    if not file_sequence.padding() and is_image and len(frame_set) > 1:
        file_sequence.setPadding("#")

    # we should have a range by now, set it in the sequence
    if (is_image and not is_single_image) or file_sequence.padding():
        file_sequence.setFrameSet(frame_set)

    return file_sequence, start, end, is_sequence, is_image



"""Sound related classes (Clip and Project sounds)."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import Self

from pytvpaint import george, utils
from pytvpaint.utils import (
    CanMakeCurrent,
    Removable,
    refreshed_property,
    set_as_current,
)

if TYPE_CHECKING:
    from pytvpaint.clip import Clip
    from pytvpaint.project import Project

P = TypeVar("P", bound=CanMakeCurrent)


class BaseSound(Removable, ABC, Generic[P]):
    """Base sound class for project and clip sounds."""

    def __init__(
        self,
        track_index: int,
        parent: P,
    ) -> None:
        super().__init__()
        self._parent = parent
        self._data: george.TVPSound = self._info(self._parent.id, track_index)

    @staticmethod
    @abstractmethod
    def _new(sound_path: Path) -> None:
        """Create a new sound."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def _info(parent_id: str | int, track_index: int) -> george.TVPSound:
        """Get the sound data based on the track index."""
        raise NotImplementedError()

    @abstractmethod
    def _adjust(self, **kwargs: Any) -> None:
        """Modify the sound data."""
        raise NotImplementedError()

    @abstractmethod
    def remove(self) -> None:
        """Remove the sound."""
        raise NotImplementedError()

    @abstractmethod
    def reload(self) -> None:
        """Reload the sound element from source."""
        raise NotImplementedError()

    @classmethod
    def iter_sounds_data(cls, parent_id: str | int) -> Iterator[george.TVPSound]:
        """Iterator over the sound's data."""
        return utils.position_generator(
            lambda track_index: cls._info(parent_id, track_index)
        )

    def make_current(self) -> None:
        """Make the parent object current (there's no way to make a sound current)."""
        raise NotImplementedError()

    @classmethod
    def new(cls: type[Self], sound_path: Path | str, parent: P) -> Self:
        """Create a new sound from the sound path."""
        parent.make_current()
        cls._new(Path(sound_path))
        last_index = len(list(cls.iter_sounds_data(parent.id))) - 1
        return cls(last_index, parent)

    def refresh(self) -> None:
        """Refreshes the sound data."""
        super().refresh()
        if not self.refresh_on_call and self._data:
            return
        self._data = self._info(self.parent.id, self.track_index)

    @property
    def track_index(self) -> int:
        """Get the soundtrack index in the sound stack.

        Raises:
            ValueError: if sound object no longer exists
        """
        # Recomputes the track_index each time because some track
        # can be deleted in the meantime
        for index, data in enumerate(self.iter_sounds_data(self._parent.id)):
            if data == self._data:
                return index
        raise ValueError("The sound doesn't exist anymore")

    @refreshed_property
    def offset(self) -> float:
        """The sound offset in the timeline."""
        return self._data.offset

    @offset.setter
    def offset(self, value: float) -> None:
        self._adjust(offset=value)

    @refreshed_property
    def volume(self) -> float:
        """The volume of the sound."""
        return self._data.volume

    @volume.setter
    @set_as_current
    def volume(self, value: float) -> None:
        self._adjust(volume=value)

    @refreshed_property
    def is_muted(self) -> bool:
        """Returns `True` if the sound is muted."""
        return self._data.mute

    @is_muted.setter
    @set_as_current
    def is_muted(self, value: bool) -> None:
        self._adjust(mute=value)

    @refreshed_property
    def fade_in_start(self) -> float:
        """The fade in start time."""
        return self._data.fade_in_start

    @fade_in_start.setter
    @set_as_current
    def fade_in_start(self, value: float) -> None:
        self._adjust(fade_in_start=value)

    @refreshed_property
    def fade_in_stop(self) -> float:
        """The fade in stop time."""
        return self._data.fade_in_stop

    @fade_in_stop.setter
    @set_as_current
    def fade_in_stop(self, value: float) -> None:
        self._adjust(fade_in_stop=value)

    @refreshed_property
    def fade_out_start(self) -> float:
        """The fade out start time."""
        return self._data.fade_out_start

    @fade_out_start.setter
    @set_as_current
    def fade_out_start(self, value: float) -> None:
        self._adjust(fade_out_start=value)

    @refreshed_property
    def fade_out_stop(self) -> float:
        """The fade out stop time."""
        return self._data.fade_out_stop

    @fade_out_stop.setter
    @set_as_current
    def fade_out_stop(self, value: float) -> None:
        self._adjust(fade_out_stop=value)

    @refreshed_property
    def path(self) -> Path:
        """The sound source file path."""
        return self._data.path

    @refreshed_property
    def sound_in(self) -> float:
        """The start time of the sound."""
        return self._data.sound_in

    @refreshed_property
    def sound_out(self) -> float:
        """The end time of the sound."""
        return self._data.sound_out

    @refreshed_property
    def color_index(self) -> int:
        """The sound color index."""
        return self._data.color_index

    @color_index.setter
    @set_as_current
    def color_index(self, value: int) -> None:
        self._adjust(color_index=value)


class ClipSound(BaseSound["Clip"]):
    """A clip sound."""

    def __init__(
        self,
        track_index: int,
        clip: Clip | None = None,
    ) -> None:
        from pytvpaint.clip import Clip

        clip = clip or Clip.current_clip()
        super().__init__(track_index, clip)

    def __eq__(self, other: object) -> bool:
        """Two clip sounds are equal if their track index is the same."""
        if not isinstance(other, ClipSound):
            return NotImplemented
        return self.track_index == other.track_index

    @staticmethod
    def _info(parent_id: str | int, track_index: int) -> george.TVPSound:
        return george.tv_sound_clip_info(int(parent_id), track_index)

    @set_as_current
    def _adjust(self, **kwargs: Any) -> None:
        return george.tv_sound_clip_adjust(self.track_index, **kwargs)

    @staticmethod
    def _new(sound_path: Path) -> None:
        george.tv_sound_clip_new(sound_path)

    @property
    def clip(self) -> Clip:
        """The sound's clip."""
        return self._parent

    def make_current(self) -> None:
        """Makes the sound clip current."""
        self.clip.make_current()

    @set_as_current
    def remove(self) -> None:
        """Remove the sound from the clip."""
        george.tv_sound_clip_remove(self.track_index)
        self.mark_removed()

    @set_as_current
    def reload(self) -> None:
        """Reload the sound from file."""
        george.tv_sound_clip_reload(self._parent.id, self.track_index)


class ProjectSound(BaseSound["Project"]):
    """A Project sound."""

    def __init__(
        self,
        track_index: int,
        project: Project | None = None,
    ) -> None:
        from pytvpaint.project import Project

        project = project or Project.current_project()
        super().__init__(track_index, project)

    def __eq__(self, other: object) -> bool:
        """Two project sounds are equal if their track index is the same."""
        if not isinstance(other, ProjectSound):
            return NotImplemented
        return self.track_index == other.track_index

    @staticmethod
    def _new(sound_path: Path) -> None:
        george.tv_sound_project_new(sound_path)

    @staticmethod
    def _info(parent_id: str | int, track_index: int) -> george.TVPSound:
        return george.tv_sound_project_info(str(parent_id), track_index)

    @set_as_current
    def _adjust(self, **kwargs: Any) -> None:
        return george.tv_sound_project_adjust(self.track_index, **kwargs)

    @property
    def project(self) -> Project:
        """The sound's project."""
        return self._parent

    @set_as_current
    def remove(self) -> None:
        """Remove the sound from the project."""
        george.tv_sound_project_remove(self.track_index)
        self.mark_removed()

    def reload(self) -> None:
        """Reload the sound from file."""
        george.tv_sound_project_reload(self.project.id, self.track_index)

    def make_current(self) -> None:
        """Makes the sound project current."""
        self.project.make_current()



"""Camera related George functions."""

from __future__ import annotations

from dataclasses import dataclass

from pytvpaint.george.client import send_cmd
from pytvpaint.george.client.parse import (
    tv_parse_list,
    validate_args_list,
)
from pytvpaint.george.grg_base import FieldOrder, GrgErrorValue


@dataclass(frozen=True)
class TVPCamera:
    """TVPaint camera info values."""

    width: int
    height: int
    field_order: FieldOrder
    frame_rate: float
    pixel_aspect_ratio: float
    anti_aliasing: int


@dataclass(frozen=True)
class TVPCameraPoint:
    """camera 2D point info."""

    x: float
    y: float
    angle: float
    scale: float


def tv_camera_info_get() -> TVPCamera:
    """Get the information of the camera."""
    return TVPCamera(**tv_parse_list(send_cmd("tv_CameraInfo"), with_fields=TVPCamera))


def tv_camera_info_set(
    width: int | None = None,
    height: int | None = None,
    field_order: FieldOrder | None = None,
    frame_rate: float | None = None,
    pixel_aspect_ratio: float | None = None,
) -> TVPCamera:
    """Set the information of the camera."""
    optional_args = [
        (width, height),
        field_order.value if field_order else None,
        frame_rate,
        pixel_aspect_ratio,
    ]

    args = validate_args_list(optional_args)

    result = send_cmd("tv_CameraInfo", *args)
    return TVPCamera(**tv_parse_list(result, with_fields=TVPCamera))


def tv_camera_enum_points(index: int) -> TVPCameraPoint:
    """Get the position/angle/scale values of the n-th point of the camera path."""
    res = send_cmd("tv_CameraEnumPoints", index, error_values=[GrgErrorValue.NONE])
    return TVPCameraPoint(**tv_parse_list(res, with_fields=TVPCameraPoint))


def tv_camera_interpolation(position: float) -> TVPCameraPoint:
    """Get the position/angle/scale values at the given position on the camera path (between 0 and 1)."""
    res = tv_parse_list(
        send_cmd("tv_CameraInterpolation", position),
        with_fields=TVPCameraPoint,
    )
    return TVPCameraPoint(**res)


def tv_camera_insert_point(
    index: int,
    x: float,
    y: float,
    angle: float,
    scale: float,
) -> None:
    """Add a point to the camera path *before* the given index."""
    send_cmd("tv_CameraInsertPoint", index, x, y, angle, scale)


def tv_camera_remove_point(index: int) -> None:
    """Remove a point at the given index."""
    send_cmd("tv_CameraRemovePoint", index)


def tv_camera_set_point(
    index: int,
    x: float,
    y: float,
    angle: float,
    scale: float,
) -> None:
    """Set position/angle/scale value of a point at the given index and make it current."""
    send_cmd("tv_CameraSetPoint", index, x, y, angle, scale)



"""Clip object class module."""

from __future__ import annotations

from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING

from fileseq.filesequence import FileSequence

from pytvpaint import george, utils
from pytvpaint.camera import Camera
from pytvpaint.layer import Layer, LayerColor
from pytvpaint.sound import ClipSound
from pytvpaint.utils import (
    Removable,
    Renderable,
    refreshed_property,
    set_as_current,
)

if TYPE_CHECKING:
    from pytvpaint.project import Project
    from pytvpaint.scene import Scene


class Clip(Removable, Renderable):
    """A Clip is a container for layers and is part of a Scene."""

    def __init__(
        self,
        clip_id: int,
        project: Project | None = None,
    ) -> None:
        """Constructs a Clip from an existing TVPaint clip (giving its id).

        Note:
            You should use `Clip.new` to create a new clip

        Args:
            clip_id: an existing clip id
            project: the project or the current one if None
        """
        from pytvpaint.project import Project

        super().__init__()
        self._id = clip_id
        self._project = project or Project.current_project()
        self._data = george.tv_clip_info(self.id)

    def __repr__(self) -> str:
        """The string representation of the clip."""
        return f"Clip({self.name})<id:{self.id}>"

    def __eq__(self, other: object) -> bool:
        """Two clips are equal if their id is the same."""
        if not isinstance(other, Clip):
            return NotImplemented
        return self.id == other.id

    @classmethod
    def new(
        cls,
        name: str,
        scene: Scene | None = None,
        project: Project | None = None,
    ) -> Clip:
        """Creates a new clip.

        Args:
            name: the new clip name
            scene: the scene or the current one if None. Defaults to None.
            project: the project or the current one if None. Defaults to None.

        Returns:
            Clip: the newly created clip
        """
        from pytvpaint.project import Project

        project = project or Project.current_project()
        project.make_current()

        scene = scene or project.current_scene
        scene.make_current()

        name = utils.get_unique_name(project.clip_names, name)
        george.tv_clip_new(name)

        return Clip.current_clip()

    def refresh(self) -> None:
        """Refreshes the clip data."""
        super().refresh()
        if not self.refresh_on_call and self._data:
            return
        self._data = george.tv_clip_info(self._id)

    def make_current(self) -> None:
        """Make the clip the current one."""
        if george.tv_clip_current_id() == self.id:
            return
        george.tv_clip_select(self.id)

    @property
    def id(self) -> int:
        """The clip id."""
        return self._id

    @property
    def project(self) -> Project:
        """The clip's project."""
        return self._project

    @property
    def scene(self) -> Scene:
        """The clip's scene.

        Raises:
            ValueError: if clip cannot be found in the project
        """
        for scene in self.project.scenes:
            for other_clip in scene.clips:
                if other_clip == self:
                    self.make_current()
                    return scene
        raise Exception("The clip's scene doesn't exist")

    @scene.setter
    def scene(self, value: Scene) -> None:
        """Moves the clip to another scene."""
        george.tv_clip_move(self.id, value.id, self.position)

    @property
    def camera(self) -> Camera:
        """The clip camera."""
        return Camera(clip=self)

    @property
    def position(self) -> int:
        """The position of the clip in the scene.

        Raises:
            ValueError: if clip cannot be found in the project
        """
        for pos, clip_id in enumerate(self.scene.clip_ids):
            if clip_id == self.id:
                return pos
        raise ValueError(f"No clip found open with id : {self.id}")

    @position.setter
    def position(self, value: int) -> None:
        """Set the position of the clip in the scene."""
        value = max(0, value)
        george.tv_clip_move(self.id, self.scene.id, value)

    @property
    @set_as_current
    def name(self) -> str:
        """The clip name."""
        return george.tv_clip_name_get(self.id)

    @name.setter
    def name(self, value: str) -> None:
        """Set the clip name."""
        if self.name == value:
            return
        value = utils.get_unique_name(self.project.clip_names, value)
        george.tv_clip_name_set(self.id, value)

    @refreshed_property
    def start(self) -> int:
        """The start frame of the clip."""
        return self._data.first_frame + self.project.start_frame

    @refreshed_property
    def end(self) -> int:
        """The end frame of the clip."""
        return self._data.last_frame + self.project.start_frame

    @property
    def duration(self) -> int:
        """The duration of the clip in frames. Takes into account the mark in/out of the clip."""
        return (self.mark_out or self.end) - (self.mark_in or self.start) + 1

    @refreshed_property
    def timeline_start(self) -> int:
        """The start frame of the clip relative to the project's timeline."""
        # get clip real start in project timeline
        clip_real_start = 0
        for clip in self.project.clips:
            if clip == self:
                break
            clip_start = clip.mark_in or clip.start
            clip_end = clip.mark_out or clip.end
            clip_duration = (clip_end - clip_start) + 1
            clip_real_start += clip_duration

        return clip_real_start + self.project.start_frame

    @refreshed_property
    def timeline_end(self) -> int:
        """The end frame of the clip relative to the project's timeline."""
        clip_start = self.mark_in or self.start
        clip_end = self.mark_out or self.end
        clip_duration = clip_end - clip_start
        return self.timeline_start + clip_duration

    @refreshed_property
    def frame_count(self) -> int:
        """The clip's frame count."""
        return self._data.frame_count

    @property
    def is_selected(self) -> bool:
        """Returns True if the clip is selected."""
        return george.tv_clip_selection_get(self.id)

    @is_selected.setter
    def is_selected(self, value: bool) -> None:
        """Select or deselect the clip."""
        george.tv_clip_selection_set(self.id, value)

    @property
    def is_visible(self) -> bool:
        """Returns True if the clip is visible."""
        return not george.tv_clip_hidden_get(self.id)

    @is_visible.setter
    def is_visible(self, value: bool) -> None:
        """Show or hide the clip."""
        george.tv_clip_hidden_set(self.id, not value)

    @property
    def color_index(self) -> int:
        """Get the clip color index."""
        return george.tv_clip_color_get(self.id)

    @color_index.setter
    def color_index(self, value: int) -> None:
        """Set the clip color index."""
        george.tv_clip_color_set(self.id, value)

    @property
    def action_text(self) -> str:
        """Get the action text of the clip."""
        return george.tv_clip_action_get(self.id)

    @action_text.setter
    def action_text(self, value: str) -> None:
        """Set the action text of the clip."""
        george.tv_clip_action_set(self.id, value)

    @property
    def dialog_text(self) -> str:
        """Get the dialog text of the clip."""
        return george.tv_clip_dialog_get(self.id)

    @dialog_text.setter
    def dialog_text(self, value: str) -> None:
        """Set the dialog text of the clip."""
        george.tv_clip_dialog_set(self.id, value)

    @property
    def note_text(self) -> str:
        """Get the note text of the clip."""
        return george.tv_clip_note_get(self.id)

    @note_text.setter
    def note_text(self, value: str) -> None:
        """Set the note text of the clip."""
        george.tv_clip_note_set(self.id, value)

    @refreshed_property
    def is_current(self) -> bool:
        """Returns True if the clip is the current one."""
        return Clip.current_clip_id() == self.id

    @staticmethod
    def current_clip_id() -> int:
        """Returns the id of the current clip."""
        return george.tv_clip_current_id()

    @staticmethod
    def current_clip() -> Clip:
        """Returns the current clip object."""
        from pytvpaint.project import Project

        return Clip(Clip.current_clip_id(), Project.current_project())

    @property
    @set_as_current
    def current_frame(self) -> int:
        """Returns the current frame in the clip (timeline) relative to the project start frame."""
        return george.tv_layer_image_get() + self.project.start_frame

    @current_frame.setter
    @set_as_current
    def current_frame(self, value: int) -> None:
        """Set the current frame of the clip."""
        george.tv_layer_image(value - self.project.start_frame)

    @set_as_current
    def duplicate(self) -> Clip:
        """Duplicates the clip.

        Note:
            a new unique name is choosen for the duplicated clip with `get_unique_name`.
        """
        george.tv_clip_duplicate(self.id)
        new_clip = self.project.current_clip
        new_clip.name = utils.get_unique_name(self.project.clip_names, new_clip.name)
        return new_clip

    def remove(self) -> None:
        """Removes the clip.

        Warning:
            the instance is not usable after that call because it's marked as removed.
        """
        george.tv_clip_close(self.id)
        self.mark_removed()

    @property
    @set_as_current
    def layer_ids(self) -> Iterator[int]:
        """Iterator over the layer ids."""
        return utils.position_generator(lambda pos: george.tv_layer_get_id(pos))

    @property
    def layers(self) -> Iterator[Layer]:
        """Iterator over the clip's layers."""
        from pytvpaint.layer import Layer

        for layer_id in self.layer_ids:
            yield Layer(layer_id, clip=self)

    @property
    @set_as_current
    def layer_names(self) -> Iterator[str]:
        """Iterator over the clip's layer names."""
        for layer in self.layers:
            yield layer.name

    @property
    def current_layer(self) -> Layer:
        """Get the current layer in the clip.

        Raises:
            ValueError: if clip cannot be found in the project
        """
        for layer in self.layers:
            if layer.is_current:
                return layer
        raise Exception("Couldn't find a current layer")

    def get_layer(
        self,
        by_id: int | None = None,
        by_name: str | None = None,
    ) -> Layer | None:
        """Get a specific layer by id or name."""
        return utils.get_tvp_element(self.layers, by_id, by_name)

    @set_as_current
    def add_layer(self, layer_name: str) -> Layer:
        """Add a new layer in the layer stack."""
        return Layer.new(name=layer_name, clip=self)

    @property
    def selected_layers(self) -> Iterator[Layer]:
        """Iterator over the selected layers."""
        yield from (layer for layer in self.layers if layer.is_selected)

    @property
    def visible_layers(self) -> Iterator[Layer]:
        """Iterator over the visible layers."""
        yield from (layer for layer in self.layers if layer.is_visible)

    @set_as_current
    @george.undoable
    def load_media(
        self,
        media_path: Path,
        start_count: tuple[int, int] | None = None,
        stretch: bool = False,
        time_stretch: bool = False,
        preload: bool = False,
        with_name: str = "",
        field_order: george.FieldOrder = george.FieldOrder.LOWER,
    ) -> Layer:
        """Loads a media (single frame, video, ...) into a new Layer in the clip.

        Args:
            media_path: the path of the media. If it's a file sequence, give the path of the first image.
            start_count: the start and number of image of sequence to load. Defaults to None.
            stretch: Stretch each image to the size of the layer. Defaults to None.
            time_stretch: Once loaded, the layer will have a new number of image corresponding to the project framerate. Defaults to None.
            preload: Load all the images in memory, no more reference on the files. Defaults to None.
            with_name: the name of the new layer
            field_order: the field order. Defaults to None.

        Returns:
            Layer: the new layer created
        """
        media_path = Path(media_path)

        george.tv_load_sequence(
            media_path,
            start_count,
            field_order,
            stretch,
            time_stretch,
            preload,
        )

        new_layer = Layer.current_layer()
        new_layer.name = with_name or media_path.stem

        return new_layer

    def _validate_range(self, start: int, end: int) -> None:
        clip_start = self.start
        clip_end = self.end
        clip_mark_in = self.mark_in
        clip_mark_out = self.mark_out

        clip_full_range = (
            (clip_mark_in if clip_mark_in else clip_start),
            (clip_mark_out if clip_mark_out else clip_end),
        )
        if start < clip_full_range[0] or end > clip_full_range[1]:
            raise ValueError(
                f"Render ({start}-{end}) not in clip range ({clip_full_range})"
            )

    def _get_real_range(self, start: int, end: int) -> tuple[int, int]:
        # get project start to get real values
        project_start_frame = self.project.start_frame
        # get clip real start in project timeline
        clip_real_start = self.timeline_start - project_start_frame
        # get real mark_in since we'll also need to subtract it from the range
        real_mark_in = (self.mark_in - project_start_frame) if self.mark_in else 0

        start = (start - project_start_frame - real_mark_in) + clip_real_start
        end = (end - project_start_frame - real_mark_in) + clip_real_start

        # clamp values to clip start
        start = max(clip_real_start, start)
        end = max(clip_real_start, end)
        return start, end

    @set_as_current
    def render(
        self,
        output_path: Path | str | FileSequence,
        start: int | None = None,
        end: int | None = None,
        use_camera: bool = False,
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Render the clip to a single frame or frame sequence or movie.

        Args:
            output_path: a single file or file sequence pattern
            start: the start frame to render or the mark in or the clip's start if None. Defaults to None.
            end: the end frame to render or the mark out or the clip's end if None. Defaults to None.
            use_camera: use the camera for rendering, otherwise render the whole canvas. Defaults to False.
            layer_selection: list of layers to render, if None render all of them. Defaults to None.
            alpha_mode: the alpha mode for rendering. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the background mode for rendering. Defaults to None.
            format_opts: custom format options. Defaults to None.

        Raises:
            ValueError: if requested range (start-end) not in clip range/bounds
            ValueError: if output is a movie, and it's duration is equal to 1 frame
            FileNotFoundError: if the render failed and no files were found on disk or missing frames

        Note:
            This functions uses the clip's range as a basis (start-end). This  is different from a project range, which
            uses the project timeline. For more details on the differences in frame ranges and the timeline in TVPaint,
            please check the `Usage/Rendering` section of the documentation.

        Warning:
            Even tough pytvpaint does a pretty good job of correcting the frame ranges for rendering, we're still
            encountering some weird edge cases where TVPaint will consider the range invalid for seemingly no reason.
        """
        default_start = self.mark_in or self.start
        default_end = self.mark_out or self.end

        self._render(
            output_path,
            default_start,
            default_end,
            start,
            end,
            use_camera,
            layer_selection=layer_selection,
            alpha_mode=alpha_mode,
            background_mode=background_mode,
            format_opts=format_opts,
        )

    @set_as_current
    def export_tvp(self, export_path: Path | str) -> None:
        """Exports the clip in .tvp format which can be imported as a project in TVPaint.

        Raises:
            ValueError: if output extension is not (.tvp)
            FileNotFoundError: if the render failed and no files were found on disk
        """
        export_path = Path(export_path)

        if export_path.suffix != ".tvp":
            raise ValueError("The file extension must be .tvp")

        export_path.parent.mkdir(exist_ok=True, parents=True)
        george.tv_save_clip(export_path)

        if not export_path.exists():
            raise FileNotFoundError(
                f"Could not find output at : {export_path.as_posix()}"
            )

    @set_as_current
    def export_json(
        self,
        export_path: Path | str,
        save_format: george.SaveFormat,
        folder_pattern: str = r"[%3li] %ln",
        file_pattern: str = r"[%3ii] %ln",
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
        all_images: bool = False,
        ignore_duplicates: bool = False,
    ) -> None:
        """Exports the instances (or all the images) of layers in the clip and a JSON file describing the structure of that clip.

        Args:
            export_path: the JSON export path
            save_format: file format to use for rendering
            folder_pattern: the folder name pattern (%li: layer index, %ln: layer name, %fi: file index (added in 11.0.8)). Defaults to None.
            file_pattern: the file name pattern (%li: layer index, %ln: layer name, %ii: image index, %in: image name, %fi: file index (added in 11.0.8)). Defaults to None.
            layer_selection: list of layers to render or all if None. Defaults to None.
            alpha_mode: the export alpha mode. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the export background mode. Defaults to None.
            format_opts: custom format options. Defaults to None.
            all_images: export all images (not only the instances). Defaults to False.
            ignore_duplicates: Ignore duplicates images. Defaults to None.

        Raises:
            FileNotFoundError: if the export failed and no files were found on disk
        """
        export_path = Path(export_path)
        export_path.parent.mkdir(exist_ok=True, parents=True)

        fill_background = bool(
            background_mode not in [None, george.BackgroundMode.NONE]
        )

        with utils.render_context(
            alpha_mode, background_mode, save_format, format_opts, layer_selection
        ):
            george.tv_clip_save_structure_json(
                export_path,
                save_format,
                fill_background=fill_background,
                folder_pattern=folder_pattern,
                file_pattern=file_pattern,
                visible_layers_only=True,
                all_images=all_images,
                ignore_duplicates=ignore_duplicates,
            )

        if not export_path.exists():
            raise FileNotFoundError(
                f"Could not find output at : {export_path.as_posix()}"
            )

    @set_as_current
    def export_psd(
        self,
        export_path: Path | str,
        mode: george.PSDSaveMode,
        start: int | None = None,
        end: int | None = None,
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Save the current clip as a PSD.

        Args:
            export_path: the PSD save path
            mode: whether to save all the images, only the instances or inside the markin
            start: the start frame. Defaults to None.
            end: the end frame. Defaults to None.
            layer_selection: layers to render. Defaults to None (render all the layers).
            alpha_mode: the alpha save mode. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the export background mode. Defaults to None.
            format_opts: custom format options. Defaults to None.

        Raises:
            FileNotFoundError: if the export failed and no files were found on disk
        """
        start = start or self.mark_in or self.start
        end = end or self.mark_out or self.end

        export_path = Path(export_path)
        image = start if mode == george.PSDSaveMode.IMAGE else None

        with utils.render_context(
            alpha_mode,
            background_mode,
            george.SaveFormat.PSD,
            format_opts,
            layer_selection,
        ):
            george.tv_clip_save_structure_psd(
                export_path,
                mode,
                image=image,
                mark_in=start,
                mark_out=end,
            )

        if mode == george.PSDSaveMode.MARKIN:
            # raises error if sequence not found
            check_path = export_path.with_suffix(f".#{export_path.suffix}").as_posix()
            assert FileSequence.findSequenceOnDisk(check_path)
        else:
            if not export_path.exists():
                raise FileNotFoundError(
                    f"Could not find output at : {export_path.as_posix()}"
                )

    @set_as_current
    def export_csv(
        self,
        export_path: Path | str,
        save_format: george.SaveFormat,
        all_images: bool = False,
        exposure_label: str = "",
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Save the current clip as a CSV.

        Args:
            export_path: the .csv export path
            save_format: file format to use for rendering
            all_images: export all images or only instances. Defaults to None.
            exposure_label: give a label when the image is an exposure. Defaults to None.
            layer_selection: layers to render. Defaults to None (render all the layers).
            alpha_mode: the alpha save mode. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the export background mode. Defaults to None.
            format_opts: custom format options. Defaults to None.

        Raises:
            ValueError: if the extension is not .csv
            FileNotFoundError: if the render failed and no files were found on disk
        """
        export_path = Path(export_path)

        if export_path.suffix != ".csv":
            raise ValueError("Export path must have .csv extension")

        with utils.render_context(
            alpha_mode, background_mode, save_format, format_opts, layer_selection
        ):
            george.tv_clip_save_structure_csv(export_path, all_images, exposure_label)

        if not export_path.exists():
            raise FileNotFoundError(
                f"Could not find output at : {export_path.as_posix()}"
            )

    @set_as_current
    def export_sprites(
        self,
        export_path: Path | str,
        layout: george.SpriteLayout | None = None,
        space: int = 0,
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Save the current clip as sprites in one image.

        Args:
            export_path (Path | str): _description_
            layout: the sprite layout. Defaults to None.
            space: the space between each sprite in the image. Defaults to None.
            layer_selection: layers to render. Defaults to None (render all the layers).
            alpha_mode: the alpha save mode. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the export background mode. Defaults to None.
            format_opts: custom format options. Defaults to None.

        Raises:
            FileNotFoundError: if the export failed and no files were found on disk
        """
        export_path = Path(export_path)
        save_format = george.SaveFormat.from_extension(export_path.suffix)

        with utils.render_context(
            alpha_mode, background_mode, save_format, format_opts, layer_selection
        ):
            george.tv_clip_save_structure_sprite(export_path, layout, space)

        if not export_path.exists():
            raise FileNotFoundError(
                f"Could not find output at : {export_path.as_posix()}"
            )

    @set_as_current
    def export_flix(
        self,
        export_path: Path | str,
        start: int | None = None,
        end: int | None = None,
        import_parameters: str = "",
        file_parameters: str = "",
        send: bool = False,
        layer_selection: list[Layer] | None = None,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Save the current clip for Flix.

        Args:
            export_path: the .xml export path
            start: the start frame. Defaults to None.
            end: the end frame. Defaults to None.
            import_parameters: the attribute(s) of the global <flixImport> tag (waitForSource/...). Defaults to None.
            file_parameters: the attribute(s) of each <image> (file) tag (dialogue/...). Defaults to None.
            send: open a browser with the prefilled url. Defaults to None.
            layer_selection: layers to render. Defaults to None (render all the layers).
            alpha_mode: the alpha save mode. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the export background mode. Defaults to None.
            format_opts: custom format options. Defaults to None.

        Raises:
            ValueError: if the extension is not .xml
            FileNotFoundError: if the export failed and no files were found on disk
        """
        export_path = Path(export_path)

        if export_path.suffix != ".xml":
            raise ValueError("Export path must have .xml extension")

        original_file = self.project.path
        import_parameters = (
            import_parameters
            or 'waitForSource="1" multipleSetups="1" replaceSelection="0"'
        )

        # The project needs to be saved
        self.project.save()

        # save alpha mode and save format values
        with utils.render_context(
            alpha_mode, background_mode, None, format_opts, layer_selection
        ):
            george.tv_clip_save_structure_flix(
                export_path,
                start,
                end,
                import_parameters,
                file_parameters,
                send,
                original_file,
            )

        if not export_path.exists():
            raise FileNotFoundError(
                f"Could not find output at : {export_path.as_posix()}"
            )

    @property
    @set_as_current
    def mark_in(self) -> int | None:
        """Get the mark in of the clip."""
        frame, mark_action = george.tv_mark_in_get(reference=george.MarkReference.CLIP)
        if mark_action == george.MarkAction.CLEAR:
            return None
        return frame + self.project.start_frame

    @mark_in.setter
    @set_as_current
    def mark_in(self, value: int | None) -> None:
        """Set the mark int value of the clip or None to clear it."""
        if value is None:
            action = george.MarkAction.CLEAR
            value = self.mark_in or 0
        else:
            action = george.MarkAction.SET
            value = value

        frame = value - self.project.start_frame
        george.tv_mark_in_set(
            reference=george.MarkReference.CLIP, frame=frame, action=action
        )

    @property
    @set_as_current
    def mark_out(self) -> int | None:
        """Get the mark out of the clip."""
        frame, mark_action = george.tv_mark_out_get(reference=george.MarkReference.CLIP)
        if mark_action == george.MarkAction.CLEAR:
            return None
        return frame + self.project.start_frame

    @mark_out.setter
    @set_as_current
    def mark_out(self, value: int | None) -> None:
        """Set the mark in of the clip or None to clear it."""
        if value is None:
            action = george.MarkAction.CLEAR
            value = self.mark_out or 0
        else:
            action = george.MarkAction.SET
            value = value

        frame = value - self.project.start_frame
        george.tv_mark_out_set(
            reference=george.MarkReference.CLIP,
            frame=frame,
            action=action,
        )

    @property
    def layer_colors(self) -> Iterator[LayerColor]:
        """Iterator over the layer colors."""
        for color_index in range(26):
            yield LayerColor(color_index=color_index, clip=self)

    def set_layer_color(self, layer_color: LayerColor) -> None:
        """Set the layer color at the provided index.

        Args:
            layer_color: the layer color instance.
        """
        george.tv_layer_color_set_color(
            self.id, layer_color.index, layer_color.color, layer_color.name
        )

    def get_layer_color(
        self,
        by_index: int | None = None,
        by_name: str | None = None,
    ) -> LayerColor | None:
        """Get a layer color by index or name.

        Raises:
            ValueError: if none of the arguments `by_index` and `by_name` where provided
        """
        if not by_index and by_name:
            raise ValueError(
                "At least one value (by_index or by_name) must be provided"
            )

        if by_index is not None:
            return next(c for i, c in enumerate(self.layer_colors) if i == by_index)

        try:
            return next(c for c in self.layer_colors if c.name == by_name)
        except StopIteration:
            return None

    @property
    def bookmarks(self) -> Iterator[int]:
        """Iterator over the clip bookmarks."""
        bookmarks_iter = utils.position_generator(
            lambda pos: george.tv_bookmarks_enum(pos)
        )
        project_start_frame = self.project.start_frame
        return (frame + project_start_frame for frame in bookmarks_iter)

    def add_bookmark(self, frame: int) -> None:
        """Add a bookmark at that frame."""
        george.tv_bookmark_set(frame - self.project.start_frame)

    def remove_bookmark(self, frame: int) -> None:
        """Remove a bookmark at that frame."""
        george.tv_bookmark_clear(frame - self.project.start_frame)

    def clear_bookmarks(self) -> None:
        """Remove all the bookmarks."""
        bookmarks = list(self.bookmarks)
        for frame in bookmarks:
            self.remove_bookmark(frame)

    @set_as_current
    def go_to_previous_bookmark(self) -> None:
        """Go to the previous bookmarks frame."""
        george.tv_bookmark_prev()

    @set_as_current
    def go_to_next_bookmark(self) -> None:
        """Go to the next bookmarks frame."""
        george.tv_bookmark_next()

    @property
    def sounds(self) -> Iterator[ClipSound]:
        """Iterates through the clip's soundtracks."""
        sounds_data = utils.position_generator(
            lambda pos: george.tv_sound_clip_info(self.id, pos)
        )

        for track_index, _ in enumerate(sounds_data):
            yield ClipSound(track_index, clip=self)

    def get_sound(
        self,
        by_id: int | None = None,
        by_path: Path | str | None = None,
    ) -> ClipSound | None:
        """Get a clip sound by id or by path.

        Raises:
            ValueError: if sound object could not be found in clip
        """
        for sound in self.sounds:
            if (by_id and sound.id == by_id) or (by_path and sound.path == by_path):
                return sound

        return None

    def add_sound(self, sound_path: Path | str) -> ClipSound:
        """Adds a new clip soundtrack."""
        return ClipSound.new(sound_path, parent=self)



"""Export all the george members with wildcard imports (it's not recommended but practical in this case).

Because we can do `from pytvpaint.george import <any_tv_george_command>`.
"""

from pytvpaint.george.exceptions import *  # noqa: F403
from pytvpaint.george.grg_base import *  # noqa: F403
from pytvpaint.george.grg_camera import *  # noqa: F403
from pytvpaint.george.grg_clip import *  # noqa: F403
from pytvpaint.george.grg_layer import *  # noqa: F403
from pytvpaint.george.grg_project import *  # noqa: F403
from pytvpaint.george.grg_scene import *  # noqa: F403



"""Parsing functions used to handle data coming from TVPaint and also preparing arguments for them to be sent.

The two main functions are `tv_parse_dict` and `tv_parse_list` which handle the return values of George functions.

George can either return a list of values or a list of key/value pairs which are consecutive.
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import Field, fields, is_dataclass
from enum import Enum
from typing import (
    Any,
    ClassVar,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from typing_extensions import Protocol, TypeAlias


class DataclassInstance(Protocol):
    """Protocol that describes a Dataclass instance."""

    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


def tv_handle_string(s: str) -> str:
    """String handling for George arguments. It wraps the string into quotes if it has spaces.

    See an example here: https://www.tvpaint.com/doc/tvpaint-animation-11/george-commands#tv_projectnew

    Args:
        s (str): the input string

    Returns:
        the "escaped" string
    """
    if " " in s:
        return f'"{s}"'

    return s


def camel_to_pascal(s: str) -> str:
    """Convert a camel case string to pascal case.

    Example:
        `this_is_a_text -> ThisIsAText`

    Args:
        s: the input string

    Returns:
        the string in pascal case
    """
    return "".join([c.capitalize() for c in s.split("_")])


T = TypeVar("T", bound=Any)


def tv_cast_to_type(value: str, cast_type: type[T]) -> T:
    """Cast a value to the provided type using George's convention for values.

    Note:
        "1" and "on"/"ON" values are considered True when parsing a boolean

    Args:
        value: the input value
        cast_type: the type to cast to

    Raises:
        ValueError: if given an enum, and it can't find the value or the enum index is invalid

    Returns:
        the value cast to the provided type
    """
    if issubclass(cast_type, Enum):
        value = value.strip().strip('"')

        # Find all enum members that matches the value (lower case)
        matches = [m for m in cast_type if value.lower() == m.value.lower()]
        try:
            # If the unmodified value is in the enum, return that first
            return cast(T, next(m for m in matches if value == m.value))
        except StopIteration:
            # Otherwise return the first match
            if matches:
                return cast(T, matches[0])

        # It didn't work, it can be the enum index
        try:
            index = int(value)
        except ValueError:
            raise ValueError(
                f"{value} is not a valid Enum index since it can't be parsed as int"
            )

        # We get the enum member at that index
        enum_members = list(cast_type)
        if index < len(enum_members):
            return cast(T, enum_members[index])

        raise ValueError(
            f"Enum index {index} is out of bounds (max {len(enum_members) - 1})"
        )

    if get_origin(cast_type) is tuple:
        # Split by space and convert each member to the right type
        values_types = zip(value.split(" "), get_args(cast_type))
        return cast(T, tuple(tv_cast_to_type(v, t) for v, t in values_types))

    if cast_type == bool:
        return cast(T, value.lower() in ["1", "on", "true"])

    if cast_type == str:
        return cast(T, value.strip().strip('"'))

    return cast(T, cast_type(value))


FieldTypes: TypeAlias = list[tuple[str, Any]]


def _get_dataclass_fields(
    datacls: DataclassInstance | type[DataclassInstance],
) -> FieldTypes:
    """Get the dataclass key/type pairs and filter those with the "parsed" metadata.

    Args:
        datacls: input dataclass

    Returns:
        the list of key/type tuple
    """
    type_hints = get_type_hints(datacls)
    return [
        (f.name, type_hints[f.name])
        for f in fields(datacls)
        if f.metadata.get("parsed", True)
    ]


def tv_parse_dict(
    input_text: str,
    with_fields: FieldTypes | type[DataclassInstance],
) -> dict[str, Any]:
    """Parse a list of values as key value pairs returned from TVPaint commands.

    Cast the values to a provided dataclass type or list of key/types pairs.

    Args:
        input_text: the George string result
        with_fields: the field types (can be a dataclass)

    Returns:
        a dict with the values cast to the given types
    """
    # For dataclasses get the type hints and filter those with metadata
    if is_dataclass(with_fields):
        with_fields = _get_dataclass_fields(with_fields)
    else:
        # Explicitly cast because we are sure now
        with_fields = cast(FieldTypes, with_fields)

    output_dict: dict[str, Any] = {}
    search_start = 0

    for i, (field_name, field_type) in enumerate(with_fields):
        current_key_pascal = camel_to_pascal(field_name)

        # Search for the key from the end
        search_text = input_text.lower()
        try:
            start = search_text.index(current_key_pascal.lower(), search_start)
        except ValueError:
            continue

        if i < (len(with_fields) - 1):
            # Search for the next key also from the end
            next_key_pascal = camel_to_pascal(with_fields[i + 1][0])
            end = search_text.rfind(" " + next_key_pascal.lower(), search_start)
        else:
            end = len(input_text)

        # Get the "key value" substring
        key_value = input_text[start:end]

        # Extract the value
        cut_at = len(current_key_pascal) + 1
        value = key_value[cut_at:].strip()

        # Cast it to the corresponding type
        value = tv_cast_to_type(value, field_type)

        output_dict[field_name] = value
        search_start = end

    return output_dict


def tv_parse_list(
    output: str,
    with_fields: FieldTypes | type[DataclassInstance],
    unused_indices: list[int] | None = None,
) -> dict[str, Any]:
    """Parse a list of values returned from TVPaint commands.

    Cast the values to a provided dataclass type or list of key/types pairs.

    You can specify unused indices to exclude positional values from being parsed.
    This is useful because some George commands have unused return values.

    Args:
        output: the input string
        with_fields: the field types (can be a dataclass)
        unused_indices: Some George functions return positional arguments that are unused. Defaults to None.

    Returns:
        a dict with the values cast to the given types
    """
    start = 0
    current = 0
    string_open = False
    tokens: list[str] = []

    while current < len(output):
        char = output[current]
        is_quote = char == '"'
        is_space = char == " "
        last_char = current == len(output) - 1

        if (is_space and not string_open) or last_char:
            last_cut = current if not last_char else current + 1
            token = output[start:last_cut]
            if start != last_cut and token != " ":
                tokens.append(token)
            start = current = current + 1
        elif is_quote:
            if string_open:
                tokens.append(output[start:current])

            start = current = current + 1
            string_open = not string_open
        else:
            current += 1

    # Get type annotations from the dataclass fields
    if is_dataclass(with_fields):
        with_fields = _get_dataclass_fields(with_fields)
    else:
        # Explicitly cast because we are sure now
        with_fields = cast(FieldTypes, with_fields)

    # Remove any unused values
    if unused_indices:
        tokens = [t for i, t in enumerate(tokens) if i not in unused_indices]

    # Cast each token to a type and construct the dict
    tokens_dict: dict[str, Any] = {}
    for token, (field_name, field_type) in zip(tokens, with_fields):
        token = tv_cast_to_type(token, field_type)
        tokens_dict[field_name] = token

    return tokens_dict


def args_dict_to_list(args: dict[str, Any]) -> list[Any]:
    """Converts a dict of named arguments to a flat list of key/values.

    It also filters pairs with None values

    Args:
        args: dict of arguments

    Returns:
        key/values list
    """
    args_filter = {k: v for k, v in args.items() if v is not None}
    # Flatten dictionnary key value pairs with zip
    return [item for kv in args_filter.items() for item in kv]


Value = Union[int, float, str, bool, None]


def validate_args_list(optional_args: Sequence[Value | tuple[Value, ...]]) -> list[Any]:
    """Some George functions only accept a list of values and not key:value pairs, so to set the last positional argument for instance, you need to give all the previous ones.

    This function allows you to give a list of argument or key:value pairs (as tuples) and check that they are not None.

    For example, for `tv_camerainfo [<iWidth> <iHeight> [<field_order>]]`
    you can't give `[500, None, "upper"]` because `<iHeight>` is not defined.

    Args:
        optional_args: list of values or tuple of values (args block)

    Raises:
        ValueError: if not all the parameters were given

    Returns:
        the list of parameters
    """
    args: list[Any] = []

    for arg in optional_args:
        if arg is None or (isinstance(arg, tuple) and all(a is None for a in arg)):
            break

        # If it's a tuple they need to be defined
        if isinstance(arg, tuple):
            if any(a is None for a in arg):
                raise ValueError(f"You must pass all the parameters: {arg}")
            args.extend(arg)
        else:
            args.append(arg)

    return args



"""Project class."""

from __future__ import annotations

from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING

from fileseq.filesequence import FileSequence

from pytvpaint import george, utils
from pytvpaint.george.exceptions import GeorgeError
from pytvpaint.sound import ProjectSound
from pytvpaint.utils import (
    Refreshable,
    Renderable,
    refreshed_property,
    set_as_current,
)

if TYPE_CHECKING:
    from pytvpaint.clip import Clip
    from pytvpaint.scene import Scene


class Project(Refreshable, Renderable):
    """A TVPaint project is the highest object that contains everything in the data hierarchy.

    It looks like this: Project -> Scene -> Clip -> Layer -> LayerInstance
    """

    def __init__(self, project_id: str) -> None:
        super().__init__()
        self._id = project_id
        self._is_closed = False
        self._data = george.tv_project_info(self._id)

    def __repr__(self) -> str:
        """String representation of the project."""
        return f"Project({self.name})<id:{self.id}>"

    def __eq__(self, other: object) -> bool:
        """Two projects are equal if their id are the same."""
        if not isinstance(other, Project):
            return NotImplemented
        return self.id == other.id

    def refresh(self) -> None:
        """Refreshed the project data.

        Raises:
            ValueError: if project has been closed
        """
        if self._is_closed:
            msg = "Project already closed, load the project again to get data"
            raise ValueError(msg)
        if not self.refresh_on_call and self._data:
            return

        self._data = george.tv_project_info(self._id)

    @property
    def id(self) -> str:
        """The project id.

        Note:
            the id is persistent on project load/close.
        """
        return self._id

    @property
    def position(self) -> int:
        """The project's position in the project tabs.

        Raises:
            ValueError: if project cannot be found in open projects

        Note:
            the indices go from right to left in the UI
        """
        for pos, project_id in enumerate(self.open_projects_ids()):
            if project_id == self.id:
                return pos
        raise ValueError(f"No project found open with id: {self.id}")

    @property
    def is_closed(self) -> bool:
        """Returns `True` if the project is closed."""
        try:
            george.tv_project_info(self._id)
            self._is_closed = False
        except GeorgeError:
            self._is_closed = True

        return self._is_closed

    @property
    def exists(self) -> bool:
        """Checks if the project exists on disk."""
        return self.path.is_file() and self.path.suffix in [".tvpp", ".tvp"]

    @property
    def is_current(self) -> bool:
        """Returns `True` if the project is the current selected one in the UI."""
        return self.id == Project.current_project_id()

    def make_current(self) -> None:
        """Make the project the current one."""
        if self.is_current:
            return
        george.tv_project_select(self.id)

    @refreshed_property
    def path(self) -> Path:
        """The project path on disk."""
        return self._data.path

    @property
    def name(self) -> str:
        """The name of the project which is the filename without the extension."""
        return self.path.stem

    @property
    @set_as_current
    def width(self) -> int:
        """The width of the canvas."""
        return george.tv_get_width()

    @property
    @set_as_current
    def height(self) -> int:
        """The height of the canvas."""
        return george.tv_get_height()

    @set_as_current
    def resize(
        self,
        width: int,
        height: int,
        overwrite: bool = False,
        resize_opt: george.ResizeOption | None = None,
    ) -> Project:
        """Resize the current project and returns a new one.

        Args:
            width: the new width
            height: the new height
            overwrite: overwrite the original project, default is to create a new project
            resize_opt: how to resize the project

        Returns:
            the newly resized project
        """
        if (width, height) == (self.width, self.height):
            return self

        origin_position = self.position
        origin_path = self.path

        if resize_opt:
            george.tv_resize_page(width, height, resize_opt)
        else:
            george.tv_resize_project(width, height)

        # The resized project is at the same position and replaced the original one
        resized_id = george.tv_project_enum_id(origin_position)
        resized_project = Project(resized_id)

        if overwrite:
            resized_project.save(origin_path)

        return resized_project

    @property
    @set_as_current
    def fps(self) -> float:
        """The project's framerate."""
        project_fps, _ = george.tv_frame_rate_get()
        return project_fps

    @property
    @set_as_current
    def playback_fps(self) -> float:
        """The project's playback framerate."""
        _, playback_fps = george.tv_frame_rate_get()
        return playback_fps

    @set_as_current
    def set_fps(
        self,
        fps: float,
        time_stretch: bool = False,
        preview: bool = False,
    ) -> None:
        """Set the project's framerate."""
        george.tv_frame_rate_set(fps, time_stretch, preview)

    @property
    def field_order(self) -> george.FieldOrder:
        """The field order."""
        return george.tv_get_field()

    @refreshed_property
    @set_as_current
    def pixel_aspect_ratio(self) -> float:
        """The project's pixel aspect ratio."""
        return self._data.pixel_aspect_ratio

    @property
    @set_as_current
    def start_frame(self) -> int:
        """The project's start frame."""
        return george.tv_start_frame_get()

    @start_frame.setter
    @set_as_current
    def start_frame(self, value: int) -> None:
        george.tv_start_frame_set(value)

    @property
    @set_as_current
    def end_frame(self) -> int:
        """The project's end frame, meaning the last frame of the last clip in the project's timeline."""
        clips_duration = sum(clip.duration for clip in self.clips)
        return self.start_frame + clips_duration - 1

    @property
    @set_as_current
    def current_frame(self) -> int:
        """Get the current frame relative to the timeline."""
        return george.tv_project_current_frame_get() + self.start_frame

    @current_frame.setter
    @set_as_current
    def current_frame(self, value: int) -> None:
        # when setting the current frame, if it is outside the current clip's range, TVP will switch to the clip but not
        # the required frame. So we need to set it twice, one to switch the clip, and once again to set the frame
        set_twice = not (
            self.current_clip.timeline_start <= value <= self.current_clip.timeline_end
        )

        real_frame = value - self.start_frame
        george.tv_project_current_frame_set(real_frame)
        if set_twice:
            george.tv_project_current_frame_set(real_frame)

    @property
    @set_as_current
    def background_mode(self) -> george.BackgroundMode:
        """Get/Set the background mode."""
        return george.tv_background_get()[0]

    @background_mode.setter
    @set_as_current
    def background_mode(self, mode: george.BackgroundMode) -> None:
        george.tv_background_set(mode)

    @property
    @set_as_current
    def background_colors(
        self,
    ) -> tuple[george.RGBColor, george.RGBColor] | george.RGBColor | None:
        """Get/Set the background color(s).

        Returns:
            a tuple of two colors if checker, a single color if solid or None if empty
        """
        return george.tv_background_get()[1]

    @background_colors.setter
    @set_as_current
    def background_colors(
        self,
        colors: tuple[george.RGBColor, george.RGBColor] | george.RGBColor,
    ) -> None:
        george.tv_background_set(self.background_mode, colors)

    @set_as_current
    def clear_background(self) -> None:
        """Clear the background color and set it to None."""
        self.background_mode = george.BackgroundMode.NONE
        self.background_colors = (
            george.RGBColor(255, 255, 255),
            george.RGBColor(0, 0, 0),
        )

    @property
    def header_info(self) -> str:
        """The project's header info."""
        return george.tv_project_header_info_get(self.id)

    @header_info.setter
    def header_info(self, value: str) -> None:
        george.tv_project_header_info_set(self.id, value)

    @property
    def author(self) -> str:
        """The project's author info."""
        return george.tv_project_header_author_get(self.id)

    @author.setter
    def author(self, value: str) -> None:
        george.tv_project_header_author_set(self.id, value)

    @property
    def notes(self) -> str:
        """The project's notes text."""
        return george.tv_project_header_notes_get(self.id)

    @notes.setter
    def notes(self, value: str) -> None:
        george.tv_project_header_notes_set(self.id, value)

    @classmethod
    def get_project(
        cls,
        by_id: str | None = None,
        by_name: str | None = None,
    ) -> Project | None:
        """Find a project by id or by name."""
        for project in Project.open_projects():
            if (by_id and project.id == by_id) or (by_name and project.name == by_name):
                return project

        return None

    @staticmethod
    def current_scene_ids() -> Iterator[int]:
        """Yields the current project's scene ids."""
        return utils.position_generator(lambda pos: george.tv_scene_enum_id(pos))

    @property
    def current_scene(self) -> Scene:
        """Get the current scene of the project.

        Raises:
            ValueError: if scene cannot be found in project
        """
        for scene in self.scenes:
            if scene.is_current:
                return scene
        raise ValueError(f"Could not find current scene in Project {Project}")

    @property
    @set_as_current
    def scenes(self) -> Iterator[Scene]:
        """Yields the project's scenes."""
        from pytvpaint.scene import Scene

        for scene_id in self.current_scene_ids():
            yield Scene(scene_id, self)

    def get_scene(
        self,
        by_id: int | None = None,
        by_name: str | None = None,
    ) -> Scene | None:
        """Find a scene in the project by id or name."""
        for scene in self.scenes:
            if (by_id and scene.id == by_id) or (by_name and scene.name == by_name):
                return scene

        return None

    @set_as_current
    def add_scene(self) -> Scene:
        """Add a new scene in the project."""
        from pytvpaint.scene import Scene

        return Scene.new(project=self)

    @property
    @set_as_current
    def current_clip(self) -> Clip:
        """Returns the project's current clip."""
        from pytvpaint.clip import Clip

        return Clip.current_clip()

    @property
    def clips(self) -> Iterator[Clip]:
        """Iterates over all the clips in the project's scenes."""
        for scene in self.scenes:
            yield from scene.clips

    @property
    @set_as_current
    def clip_names(self) -> Iterator[str]:
        """Optimized way to get the clip names. Useful for `get_unique_name`."""
        for scene_id in self.current_scene_ids():
            clip_ids = utils.position_generator(
                lambda pos: george.tv_clip_enum_id(scene_id, pos)
            )
            for clip_id in clip_ids:
                yield george.tv_clip_name_get(clip_id)

    def get_clip(
        self,
        by_id: int | None = None,
        by_name: str | None = None,
        scene_id: int | None = None,
    ) -> Clip | None:
        """Find a clip by id, name or scene_id."""
        clips = self.clips
        if scene_id:
            selected_scene = self.get_scene(by_id=scene_id)
            clips = selected_scene.clips if selected_scene else clips

        for clip in clips:
            if (by_id and clip.id == by_id) or (by_name and clip.name == by_name):
                return clip

        return None

    def add_clip(self, clip_name: str, scene: Scene | None = None) -> Clip:
        """Add a new clip in the given scene or the current one if no scene provided."""
        scene = scene or self.current_scene
        return scene.add_clip(clip_name)

    @property
    def sounds(self) -> Iterator[ProjectSound]:
        """Iterator over the project sounds."""
        sounds_data_iter = utils.position_generator(
            lambda pos: george.tv_sound_project_info(self.id, pos)
        )

        for track_index, _ in enumerate(sounds_data_iter):
            yield ProjectSound(track_index, project=self)

    def add_sound(self, sound_path: Path | str) -> ProjectSound:
        """Add a new sound clip to the project."""
        return ProjectSound.new(sound_path, parent=self)

    def _validate_range(self, start: int, end: int) -> None:
        project_start_frame = self.start_frame
        project_end_frame = self.end_frame
        project_mark_in = self.mark_in
        project_mark_out = self.mark_out

        proj_full_range = (
            min(project_mark_in or project_start_frame, project_start_frame),
            max(project_mark_out or project_end_frame, project_end_frame),
        )
        if start < proj_full_range[0] or end > proj_full_range[1]:
            raise ValueError(
                f"Range ({start}-{end}) outside of project bounds ({proj_full_range})"
            )

    def _get_real_range(self, start: int, end: int) -> tuple[int, int]:
        project_start_frame = self.start_frame
        start -= project_start_frame
        end -= project_start_frame

        return start, end

    @set_as_current
    def render(
        self,
        output_path: Path | str | FileSequence,
        start: int | None = None,
        end: int | None = None,
        use_camera: bool = False,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Render the project to a single frame or frame sequence or movie.

        Args:
            output_path: a single file or file sequence pattern
            start: the start frame to render or the mark in or the project's start frame if None. Defaults to None.
            end: the end frame to render or the mark out or the project's end frame if None. Defaults to None.
            use_camera: use the camera for rendering, otherwise render the whole canvas. Defaults to False.
            alpha_mode: the alpha mode for rendering. Defaults to george.AlphaSaveMode.PREMULTIPLY.
            background_mode: the background mode for rendering. Defaults to george.BackgroundMode.NONE.
            format_opts: custom format options. Defaults to None.

        Raises:
            ValueError: if requested range (start-end) not in project range/bounds
            ValueError: if output is a movie, and it's duration is equal to 1 frame
            FileNotFoundError: if the render failed and no files were found on disk or missing frames

        Note:
            This functions uses the project's timeline as a basis for the range (start-end). This timeline includes all
            the project's clips and is different from a clip range. For more details on the differences in frame ranges
            and the timeline in TVPaint, please check the `Usage/Rendering` section of the documentation.

        Warning:
            Even tough pytvpaint does a pretty good job of correcting the frame ranges for rendering, we're still
            encountering some weird edge cases where TVPaint will consider the range invalid for seemingly no reason.
        """
        default_start = self.mark_in or self.start_frame
        default_end = self.mark_out or self.end_frame

        self._render(
            output_path,
            default_start,
            default_end,
            start,
            end,
            use_camera,
            layer_selection=None,
            alpha_mode=alpha_mode,
            background_mode=background_mode,
            format_opts=format_opts,
        )

    @set_as_current
    def render_clips(
        self,
        clips: list[Clip],
        output_path: Path | str | FileSequence,
        use_camera: bool = False,
        alpha_mode: george.AlphaSaveMode = george.AlphaSaveMode.PREMULTIPLY,
        background_mode: george.BackgroundMode | None = None,
        format_opts: list[str] | None = None,
    ) -> None:
        """Render sequential clips as a single output."""
        clips = sorted(clips, key=lambda c: c.position)
        start = clips[0].timeline_start
        end = clips[-1].timeline_end
        self.render(
            output_path,
            start,
            end,
            use_camera,
            alpha_mode,
            background_mode,
            format_opts,
        )

    @staticmethod
    def current_project_id() -> str:
        """Returns the current project id."""
        return george.tv_project_current_id()

    @staticmethod
    def current_project() -> Project:
        """Returns the current project."""
        return Project(project_id=Project.current_project_id())

    @staticmethod
    def open_projects_ids() -> Iterator[str]:
        """Yields the ids of the currently open projects."""
        return utils.position_generator(lambda pos: george.tv_project_enum_id(pos))

    @classmethod
    def open_projects(cls) -> Iterator[Project]:
        """Iterator over the currently open projects."""
        for project_id in Project.open_projects_ids():
            yield Project(project_id)

    @property
    @set_as_current
    def mark_in(self) -> int | None:
        """Get the project mark in or None if no mark in set."""
        frame, mark_action = george.tv_mark_in_get(
            reference=george.MarkReference.PROJECT
        )
        if mark_action == george.MarkAction.CLEAR:
            return None
        return frame + self.start_frame

    @mark_in.setter
    @set_as_current
    def mark_in(self, value: int | None) -> None:
        if value is None:
            action = george.MarkAction.CLEAR
            value = self.mark_in or 0
        else:
            action = george.MarkAction.SET
            value = value

        frame = value - self.start_frame
        george.tv_mark_in_set(
            reference=george.MarkReference.PROJECT, frame=frame, action=action
        )

    @property
    @set_as_current
    def mark_out(self) -> int | None:
        """Get the project mark out or None if no mark out set."""
        frame, mark_action = george.tv_mark_out_get(
            reference=george.MarkReference.PROJECT
        )
        if mark_action == george.MarkAction.CLEAR:
            return None
        return frame

    @mark_out.setter
    @set_as_current
    def mark_out(self, value: int | None) -> None:
        if value is None:
            action = george.MarkAction.CLEAR
            value = self.mark_out or 0
        else:
            action = george.MarkAction.SET
            value = value

        frame = value - self.start_frame
        george.tv_mark_out_set(
            reference=george.MarkReference.CLIP,
            frame=frame,
            action=action,
        )

    @classmethod
    def new(
        cls,
        project_path: Path | str,
        width: int = 1920,
        height: int = 1080,
        pixel_aspect_ratio: float = 1.0,
        frame_rate: float = 24.0,
        field_order: george.FieldOrder = george.FieldOrder.NONE,
        start_frame: int = 1,
    ) -> Project:
        """Create a new project."""
        george.tv_project_new(
            Path(project_path).resolve().as_posix(),
            width,
            height,
            pixel_aspect_ratio,
            frame_rate,
            field_order,
            start_frame,
        )
        return cls.current_project()

    @set_as_current
    def new_from_camera(self, export_path: Path | str | None = None) -> Project:
        """Create a new cropped project from the camera view."""
        cam_project_id = george.tv_project_render_camera(self.id)
        cam_project = Project(cam_project_id)

        if export_path:
            export_path = Path(export_path)
            export_path.mkdir(exist_ok=True, parents=True)
            cam_project.save(export_path)

        return cam_project

    @set_as_current
    def duplicate(self) -> Project:
        """Duplicate the project and return the new one."""
        george.tv_project_duplicate()
        duplicated = Project.current_project()
        self.make_current()
        return duplicated

    def close(self) -> None:
        """Closes the project."""
        self._is_closed = True
        george.tv_project_close(self._id)

    @classmethod
    def close_all(cls, close_tvp: bool = False) -> None:
        """Closes all open projects.

        Args:
            close_tvp: close the TVPaint instance as well
        """
        for project in list(cls.open_projects()):
            project.close()

        if close_tvp:
            george.tv_quit()

    @classmethod
    def load(cls, project_path: Path | str, silent: bool = True) -> Project:
        """Load an existing .tvpp/.tvp project or .tvpx file."""
        project_path = Path(project_path)

        # Check if project not already open, if so, return it
        for project in cls.open_projects():
            if project.path == project_path:
                return project

        george.tv_load_project(project_path, silent)
        return cls.current_project()

    def save(self, save_path: Path | str | None = None) -> None:
        """Saves the project on disk."""
        save_path = Path(save_path or self.path).resolve()
        george.tv_save_project(save_path.as_posix())

    @set_as_current
    def load_panel(self, panel_path: Path | str) -> None:
        """Load an external TVPaint panel."""
        george.tv_load_project(panel_path, silent=True)

    @set_as_current
    def load_palette(self, palette_path: Path | str) -> None:
        """Load a palette."""
        george.tv_save_palette(palette_path)

    @set_as_current
    def save_palette(self, save_path: Path | str | None = None) -> None:
        """Save a palette to the given path."""
        save_path = Path(save_path or self.path)
        george.tv_save_project(save_path)

    @set_as_current
    def save_video_dependencies(self, on_save: bool = True, now: bool = True) -> None:
        """Saves the video dependencies."""
        george.tv_project_save_video_dependencies(self.id, on_save, now)

    @set_as_current
    def save_audio_dependencies(self, on_save: bool = True) -> None:
        """Saves audio dependencies."""
        george.tv_project_save_audio_dependencies(self.id, on_save)
